@@ -1,7 +1,7 @@
 /*
  * %CopyrightBegin%
  *
- * Copyright Ericsson AB 2000-2018. All Rights Reserved.
+ * Copyright Ericsson AB 2000-2021. All Rights Reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -115,17 +115,6 @@ typedef struct {
     ErtsAlgndAsyncReadyQ *ready_queue;
 } ErtsAsyncData;
 
-#if defined(USE_VM_PROBES)
-
-/*
- * Some compilers, e.g. GCC 4.2.1 and -O3, will optimize away DTrace
- * calls if they're the last thing in the function.  :-(
- * Many thanks to Trond Norbye, via:
- * https://github.com/memcached/memcached/commit/6298b3978687530bc9d219b6ac707a1b681b2a46
- */
-static unsigned gcc_optimizer_hack = 0;
-#endif
-
 int erts_async_max_threads; /* Initialized by erl_init.c */
 int erts_async_thread_suggested_stack_size; /* Initialized by erl_init.c */
 
@@ -137,6 +126,7 @@ static void *async_main(void *);
 static ERTS_INLINE ErtsAsyncQ *
 async_q(int i)
 {
+    ASSERT(async != NULL);
     return &async->queue[i].aq;
 }
 
@@ -155,7 +145,7 @@ erts_init_async(void)
     if (erts_async_max_threads > 0) {
 	ErtsThrQInit_t qinit = ERTS_THR_Q_INIT_DEFAULT;
 	erts_thr_opts_t thr_opts = ERTS_THR_OPTS_DEFAULT_INITER;
-	char *ptr, thr_name[16];
+	char *ptr, thr_name[32];
 	size_t tot_size = 0;
 	int i;
 
@@ -209,7 +199,7 @@ erts_init_async(void)
 	for (i = 0; i < erts_async_max_threads; i++) {
 	    ErtsAsyncQ *aq = async_q(i);
 
-            erts_snprintf(thr_opts.name, 16, "async_%d", i+1);
+            erts_snprintf(thr_opts.name, sizeof(thr_name), "async_%d", i+1);
 
 	    erts_thr_create(&aq->thr_id, async_main, (void*) aq, &thr_opts);
 	}
@@ -237,10 +227,6 @@ erts_get_async_ready_queue(Uint sched_id)
 
 static ERTS_INLINE void async_add(ErtsAsync *a, ErtsAsyncQ* q)
 {
-#ifdef USE_VM_PROBES
-    int len;
-#endif
-
     if (is_internal_port(a->port)) {
 	ErtsAsyncReadyQ *arq = async_ready_q(a->sched_id);
 	a->q.prep_enq = erts_thr_q_prepare_enqueue(&arq->thr_q);
@@ -262,9 +248,6 @@ static ERTS_INLINE ErtsAsync *async_get(ErtsThrQ_t *q,
 {
     int saved_fin_deq = 0;
     ErtsThrQFinDeQ_t fin_deq;
-#ifdef USE_VM_PROBES
-    int len;
-#endif
 
     while (1) {
 	ErtsAsync *a = (ErtsAsync *) erts_thr_q_dequeue(q);
@@ -280,6 +263,7 @@ static ERTS_INLINE ErtsAsync *async_get(ErtsThrQ_t *q,
 	if (ERTS_THR_Q_DIRTY != erts_thr_q_clean(q)) {
 	    ErtsThrQFinDeQ_t tmp_fin_deq;
 
+            erts_tse_use(tse);
 	    erts_tse_reset(tse);
 
 	chk_fin_deq:
@@ -324,6 +308,7 @@ static ERTS_INLINE ErtsAsync *async_get(ErtsThrQ_t *q,
 		break;
 	    }
 
+            erts_tse_return(tse);
 	}
     }
 }
@@ -391,9 +376,10 @@ static erts_tse_t *async_thread_init(ErtsAsyncQ *aq)
     ErtsThrQInit_t qinit = ERTS_THR_Q_INIT_DEFAULT;
     erts_tse_t *tse = erts_tse_fetch();
     ERTS_DECLARE_DUMMY(Uint no);
-
     ErtsThrPrgrCallbacks callbacks;
 
+    erts_tse_return(tse);
+
     callbacks.arg = (void *) tse;
     callbacks.wakeup = async_wakeup;
     callbacks.prepare_wait = NULL;
@@ -542,7 +528,7 @@ unsigned int driver_async_port_key(ErlDrvPort port)
 **      key            pointer to secedule queue (NULL means round robin)
 **      async_invoke   function to run in thread
 **      async_data     data to pass to invoke function
-**      async_free     function for relase async_data in case of failure
+**      async_free     function for release async_data in case of failure
 */
 long driver_async(ErlDrvPort ix, unsigned int* key,
 		  void (*async_invoke)(void*), void* async_data,
