@@ -1,7 +1,7 @@
 /*
  * %CopyrightBegin%
  *
- * Copyright Ericsson AB 2001-2018. All Rights Reserved.
+ * Copyright Ericsson AB 2001-2022. All Rights Reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -172,13 +172,15 @@ dist_table_alloc(void *dep_tmpl)
     erts_rwmtx_init_opt(&dep->rwmtx, &rwmtx_opt, "dist_entry", sysname,
         ERTS_LOCK_FLAGS_CATEGORY_DISTRIBUTION);
     dep->sysname			= sysname;
+    dep->creation                       = 0; /* undefined */
     dep->cid				= NIL;
     erts_atomic_init_nob(&dep->input_handler, (erts_aint_t) NIL);
     dep->connection_id			= 0;
     dep->state				= ERTS_DE_STATE_IDLE;
-    dep->flags				= 0;
+    dep->pending_nodedown               = 0;
+    dep->suspended_nodeup               = NULL;
+    dep->dflags				= 0;
     dep->opts                           = 0;
-    dep->version			= 0;
 
     dep->mld                            = NULL;
 
@@ -201,7 +203,6 @@ dist_table_alloc(void *dep_tmpl)
     erts_port_task_handle_init(&dep->dist_cmd);
     dep->send				= NULL;
     dep->cache				= NULL;
-    dep->transcode_ctx                  = NULL;
     dep->sequences                      = NULL;
 
     /* Link in */
@@ -287,18 +288,16 @@ static ERTS_INLINE DistEntry *find_dist_entry(Eterm sysname,
         if (connected_only && is_nil(res->cid))
             res = NULL;
         else {
-            int pend_delete;
             erts_aint_t refc;
             if (inc_refc) {
                 refc = de_refc_inc_read(res, 1);
-                pend_delete = refc < 2;
+                if (refc < 2) /* Pending delete */
+                    de_refc_inc(res, 1);
             }
             else {
-                refc = de_refc_read(res, 0);
-                pend_delete = refc < 1;
+                /* Inc from 0 to 1 for pending delete */
+                erts_refc_inc_if(&ErtsDistEntry2Bin(res)->intern.refc, 0, 0);
             }
-            if (pend_delete) /* Pending delete */
-                de_refc_inc(res, 1);
         }
     }
     erts_rwmtx_runlock(&erts_dist_table_rwmtx);
@@ -396,9 +395,13 @@ erts_build_dhandle(Eterm **hpp, ErlOffHeap* ohp,
 {
     Binary *bin = ErtsDistEntry2Bin(dep);
     Eterm mref, dhandle;
+    erts_aint_t refc;
     ASSERT(bin);
     ASSERT(ERTS_MAGIC_BIN_DESTRUCTOR(bin) == erts_dist_entry_destructor);
-    mref = erts_mk_magic_ref(hpp, ohp, bin);
+    mref = erts_mk_magic_ref_get_refc(hpp, ohp, bin, &refc);
+    if (refc < 2) {
+        erts_refc_inc(&bin->intern.refc, 2); /* inc for pending delete */
+    }
     dhandle = TUPLE2(*hpp, make_small(conn_id), mref);
     *hpp += 3;
     return dhandle;
@@ -493,7 +496,7 @@ static void try_delete_dist_entry(DistEntry* dep)
 
         if (dep->state != ERTS_DE_STATE_PENDING)
             ERTS_INTERNAL_ERROR("Garbage collecting connected distribution entry");
-        erts_abort_connection_rwunlock(dep);
+        erts_abort_pending_connection_rwunlock(dep, NULL);
         report_gc_active_dist_entry(sysname, state);
     }
     else
@@ -635,7 +638,7 @@ erts_set_dist_entry_not_connected(DistEntry *dep)
     else {
         ASSERT(dep->state != ERTS_DE_STATE_IDLE);
         ASSERT(is_internal_port(dep->cid) || is_internal_pid(dep->cid));
-        if (dep->flags & DFLAG_PUBLISHED) {
+        if (dep->dflags & DFLAG_PUBLISHED) {
             ASSERT(erts_no_of_visible_dist_entries > 0);
             erts_no_of_visible_dist_entries--;
             head = &erts_visible_dist_entries;
@@ -659,7 +662,7 @@ erts_set_dist_entry_not_connected(DistEntry *dep)
 	dep->next->prev = dep->prev;
 
     dep->state = ERTS_DE_STATE_IDLE;
-    dep->flags = 0;
+    dep->dflags = 0;
     dep->opts = 0;
     dep->prev = NULL;
     dep->cid = NIL;
@@ -701,7 +704,7 @@ erts_set_dist_entry_pending(DistEntry *dep)
     erts_no_of_not_connected_dist_entries--;
 
     dep->state = ERTS_DE_STATE_PENDING;
-    dep->flags = (DFLAG_DIST_MANDATORY | DFLAG_DIST_HOPEFULLY | DFLAG_NO_MAGIC);
+    dep->dflags = (DFLAG_DIST_MANDATORY | DFLAG_DIST_HOPEFULLY | DFLAG_PENDING_CONNECT);
     dep->connection_id = (dep->connection_id + 1) & ERTS_DIST_CON_ID_MASK;
 
     ASSERT(!dep->mld);
@@ -720,7 +723,7 @@ erts_set_dist_entry_pending(DistEntry *dep)
 }
 
 void
-erts_set_dist_entry_connected(DistEntry *dep, Eterm cid, Uint flags)
+erts_set_dist_entry_connected(DistEntry *dep, Eterm cid, Uint64 flags)
 {
     erts_aint32_t set_qflgs;
 
@@ -732,6 +735,7 @@ erts_set_dist_entry_connected(DistEntry *dep, Eterm cid, Uint flags)
     ASSERT(dep != erts_this_dist_entry);
     ASSERT(is_nil(dep->cid));
     ASSERT(dep->state == ERTS_DE_STATE_PENDING);
+    ASSERT(!dep->pending_nodedown);
     ASSERT(is_internal_port(cid) || is_internal_pid(cid));
 
     if(dep->prev) {
@@ -750,7 +754,7 @@ erts_set_dist_entry_connected(DistEntry *dep, Eterm cid, Uint flags)
     erts_no_of_pending_dist_entries--;
 
     dep->state = ERTS_DE_STATE_CONNECTED;
-    dep->flags = flags & ~DFLAG_NO_MAGIC;
+    dep->dflags = flags & ~DFLAG_PENDING_CONNECT;
     dep->cid = cid;
     erts_atomic_set_nob(&dep->input_handler,
                             (erts_aint_t) cid);
@@ -891,8 +895,8 @@ ErlNode *erts_find_or_insert_node(Eterm sysname, Uint32 creation, Eterm book)
     erts_rwmtx_rlock(&erts_node_table_rwmtx);
     res = hash_get(&erts_node_table, (void *) &ne);
     if (res && res != erts_this_node) {
-	erts_aint_t refc = erts_ref_node_entry(res, 0, book);
-	if (refc < 2) /* New or pending delete */
+	erts_aint_t refc = erts_ref_node_entry(res, 1, book);
+	if (refc < 2) /* Pending delete */
             erts_ref_node_entry(res, 1, THE_NON_VALUE);
     }
     erts_rwmtx_runlock(&erts_node_table_rwmtx);
@@ -977,7 +981,7 @@ static void print_node(void *venp, void *vpndp)
 	if(pndp->sysname == NIL) {
 	    erts_print(pndp->to, pndp->to_arg, "Name: %T ", enp->sysname);
 	}
-	erts_print(pndp->to, pndp->to_arg, " %d", enp->creation);
+	erts_print(pndp->to, pndp->to_arg, " %u", enp->creation);
 #ifdef DEBUG
 	erts_print(pndp->to, pndp->to_arg, " (refc=%ld)",
 		   erts_refc_read(&enp->refc, 0));
@@ -1020,7 +1024,7 @@ void erts_print_node_info(fmtfn_t to,
 /* ----------------------------------------------------------------------- */
 
 void
-erts_set_this_node(Eterm sysname, Uint creation)
+erts_set_this_node(Eterm sysname, Uint32 creation)
 {
     ERTS_LC_ASSERT(erts_thr_progress_is_blocking());
     ASSERT(2 <= de_refc_read(erts_this_dist_entry, 2));
@@ -1271,7 +1275,9 @@ erts_get_node_and_dist_references(struct process *proc)
 
     erts_proc_unlock(proc, ERTS_PROC_LOCK_MAIN);
     erts_thr_progress_block();
-    /* No need to lock any thing since we are alone... */
+    erts_proc_lock(proc, ERTS_PROC_LOCK_MAIN);
+    /* No need to lock any thing else since we are alone...
+       ... except dirty stuff... (?) */
 
     if (references_atoms_need_init) {
 	INIT_AM(heap);
@@ -1326,7 +1332,6 @@ erts_get_node_and_dist_references(struct process *proc)
     clear_system();
 
     erts_thr_progress_unblock();
-    erts_proc_lock(proc, ERTS_PROC_LOCK_MAIN);
     return res;
 }
 
@@ -1562,6 +1567,20 @@ insert_offheap(ErlOffHeap *oh, int type, Eterm id)
     }
 }
 
+static ERTS_INLINE ErlHeapFragment *
+get_hidden_heap_frag(Eterm val)
+{
+    ErlHeapFragment *hfp;
+    if (is_immed(val))
+        hfp = NULL;
+    else {
+        ASSERT(is_CP(val));
+        hfp = (ErlHeapFragment *) cp_val(val);
+        ASSERT(hfp->alloc_size == hfp->used_size + 1);
+    }
+    return hfp;
+}
+
 static void insert_monitor_data(ErtsMonitor *mon, int type, Eterm id)
 {
     ErtsMonitorData *mdp = erts_monitor_to_data(mon);
@@ -1576,6 +1595,23 @@ static void insert_monitor_data(ErtsMonitor *mon, int type, Eterm id)
                 oh.first = mdep->uptr.ohhp;
                 insert_offheap(&oh, type, id);
             }
+            if (mdp->origin.flags & ERTS_ML_FLG_TAG) {
+                ErlHeapFragment *hfp;
+                Eterm tag_storage;
+                if (mdp->origin.flags & ERTS_ML_FLG_EXTENDED)
+                    tag_storage = ((ErtsMonitorDataExtended *) mdp)->heap[0];
+                else
+                    tag_storage = ((ErtsMonitorDataTagHeap *) mdp)->heap[0];
+                hfp = get_hidden_heap_frag(tag_storage);
+                if (hfp)
+                    insert_offheap(&hfp->off_heap, type, id);
+            }
+            if (mdp->origin.flags & ERTS_ML_FLG_SPAWN_PENDING) {
+                ErtsMonitorDataExtended *mdep = (ErtsMonitorDataExtended *) mdp;
+                ErlHeapFragment *hfp = get_hidden_heap_frag(mdep->u.name);
+                if (hfp)
+                    insert_offheap(&hfp->off_heap, type, id);
+            }
         }
     }
     mdp->origin.flags |= ERTS_ML_FLG_DBG_VISITED;
@@ -1639,18 +1675,21 @@ clear_visited_dist_monitors(DistEntry *dep)
 
 static void insert_link_data(ErtsLink *lnk, int type, Eterm id)
 {
-    ErtsLinkData *ldp = erts_link_to_data(lnk);
-    if ((ldp->a.flags & (ERTS_ML_FLG_DBG_VISITED
-                         | ERTS_ML_FLG_EXTENDED)) == ERTS_ML_FLG_EXTENDED) {
-        ErtsLinkDataExtended *ldep = (ErtsLinkDataExtended *) ldp;
-        if (ldep->ohhp) {
+    if ((lnk->flags & (ERTS_ML_FLG_DBG_VISITED
+                       | ERTS_ML_FLG_EXTENDED)) != ERTS_ML_FLG_EXTENDED) {
+        lnk->flags |= ERTS_ML_FLG_DBG_VISITED;
+    }
+    else {
+        ErtsELink *elnk = erts_link_to_elink(lnk);
+        if (elnk->ohhp) {
             ErlOffHeap oh;
             ERTS_INIT_OFF_HEAP(&oh);
-            oh.first = ldep->ohhp;
+            oh.first = elnk->ohhp;
             insert_offheap(&oh, type, id);
         }
+        elnk->ld.proc.flags |= ERTS_ML_FLG_DBG_VISITED;
+        elnk->ld.dist.flags |= ERTS_ML_FLG_DBG_VISITED;
     }
-    ldp->a.flags |= ERTS_ML_FLG_DBG_VISITED;
 }
 
 static int insert_link(ErtsLink *lnk, void *idp, Sint reds)
@@ -1662,8 +1701,13 @@ static int insert_link(ErtsLink *lnk, void *idp, Sint reds)
 
 static int clear_visited_link(ErtsLink *lnk, void *p, Sint reds)
 {
-    ErtsLinkData *ldp = erts_link_to_data(lnk);
-    ldp->a.flags &= ~ERTS_ML_FLG_DBG_VISITED;
+    if (!(lnk->flags & ERTS_ML_FLG_EXTENDED))
+        lnk->flags &= ~ERTS_ML_FLG_DBG_VISITED;
+    else {
+        ErtsELink *elnk = erts_link_to_elink(lnk);
+        elnk->ld.proc.flags &= ~ERTS_ML_FLG_DBG_VISITED;
+        elnk->ld.dist.flags &= ~ERTS_ML_FLG_DBG_VISITED;
+    }
     return 1;
 }
 
