@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 1997-2020. All Rights Reserved.
+%% Copyright Ericsson AB 1997-2021. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -49,7 +49,7 @@
 	 bad_list_to_binary/1, bad_binary_to_list/1,
 	 t_split_binary/1, bad_split/1,
 	 terms/1, terms_float/1, float_middle_endian/1,
-         b2t_used_big/1,
+         b2t_used_big/1, t2b_deterministic/1,
 	 external_size/1, t_iolist_size/1,
          t_iolist_size_huge_list/1,
          t_iolist_size_huge_bad_arg_list/1,
@@ -69,7 +69,6 @@
 	 ordering/1,unaligned_order/1,gc_test/1,
 	 bit_sized_binary_sizes/1,
 	 otp_6817/1,deep/1,
-         term2bin_tuple_fallbacks/1,
          robustness/1,otp_8117/1,
 	 otp_8180/1, trapping/1, large/1,
 	 error_after_yield/1, cmp_old_impl/1,
@@ -90,7 +89,7 @@ all() ->
      t_iolist_size_huge_list,
      t_iolist_size_huge_bad_arg_list,
      {group, iolist_size_benchmarks},
-     b2t_used_big,
+     b2t_used_big, t2b_deterministic,
      bad_binary_to_term_2, safe_binary_to_term2,
      bad_binary_to_term, bad_terms, t_hash, bad_size,
      sub_bin_copy, bad_term_to_binary, t2b_system_limit,
@@ -98,7 +97,6 @@ all() ->
      otp_5484, otp_5933,
      ordering, unaligned_order, gc_test,
      bit_sized_binary_sizes, otp_6817, otp_8117, deep,
-     term2bin_tuple_fallbacks,
      robustness, otp_8180, trapping, large,
      error_after_yield, cmp_old_impl].
 
@@ -497,9 +495,9 @@ do_t2b_system_limit() ->
                 garbage_collect(),
                 ok
         end,
-    Opts = [{args, "-pa " ++ filename:dirname(code:which(?MODULE))}],
-    {ok,Node} = test_server:start_node(?FUNCTION_NAME, slave, Opts),
-    erpc:call(Node, F).
+    {ok, Peer, Node} = ?CT_PEER(),
+    erpc:call(Node, F),
+    peer:stop(Peer).
 
 test_t2b_system_limit(HugeBin, Name, F1, F2) ->
     io:format("Testing ~p(HugeBin)~n", [Name]),
@@ -579,18 +577,28 @@ terms(Config) when is_list(Config) ->
     TestFun = fun(Term) ->
                       S = io_lib:format("~p", [Term]),
                       io:put_chars(S),
+
 		      Bin = term_to_binary(Term),
 		      case erlang:external_size(Bin) of
 			  Sz when is_integer(Sz), size(Bin) =< Sz ->
 			      ok
 		      end,
+
                       Bin1 = term_to_binary(Term, [{minor_version, 1}]),
                       case erlang:external_size(Bin1, [{minor_version, 1}]) of
                           Sz1 when is_integer(Sz1), size(Bin1) =< Sz1 ->
                               ok
                       end,
+
+                      BinDet = term_to_binary(Term, [deterministic]),
+                      case erlang:external_size(Bin) of
+                          DetSz when is_integer(DetSz), byte_size(Bin) =< DetSz ->
+                              ok
+                      end,
+
 		      Term = binary_to_term_stress(Bin),
 		      Term = binary_to_term_stress(Bin, [safe]),
+		      Term = binary_to_term_stress(BinDet),
                       Bin_sz = byte_size(Bin),
 		      {Term,Bin_sz} = binary_to_term_stress(Bin, [used]),
 
@@ -676,6 +684,156 @@ float_middle_endian(Config) when is_list(Config) ->
     <<131,70,63,240,0,0,0,0,0,0>> = term_to_binary(1.0, [{minor_version,1}]),
     1.0 = binary_to_term_stress(<<131,70,63,240,0,0,0,0,0,0>>).
 
+%% Test term_to_binary(Term, [deterministic]).
+t2b_deterministic(_Config) ->
+    _ = rand:uniform(),				%Seed generator
+    io:format("Seed: ~p", [rand:export_seed()]),
+    Map0 = t2b_deterministic_1(100, #{1 => a, 1.0 => b}),
+
+    Map1 = maps:merge(make_map(10000), Map0),
+    test_deterministic(Map1),
+
+    case total_memory() of
+        Amount when Amount > 15 ->
+            t2b_deterministic_heavy(Map0);
+        _ ->
+            {comment,"heavy tests skipped"}
+    end.
+
+t2b_deterministic_heavy(Map0) ->
+    Map1 = maps:merge(make_map(1_000_000), Map0),
+    test_deterministic(Map1),
+
+    %% If the external representation of the term to be encoded is
+    %% estimated to fit in a heap binary (64 bytes), the internal
+    %% enc_term_int() function that does the actual encoding will be
+    %% called without an context and thus is unable to trap. In the
+    %% DEBUG build of the runtime system, maps with more than 3
+    %% elements will be large maps, and depending on the keys and
+    %% values, it is possible that a large map can fit in 64
+    %% bytes. Therefore, the encoding code for deterministic maps must
+    %% cannot assume that there always exists a context.
+
+    lists:foreach(fun(N) ->
+                          Map = maps:from_list([{E,E} || E <- lists:seq(1, N)]),
+                          test_deterministic(Map)
+                  end, lists:seq(4, 16)),
+
+    %% Test that terminating a process that is encoding a huge map
+    %% will neither crash the runtime system nor leak memory. (Run the
+    %% test with memory sanitizer or valgrind to catch potential memory
+    %% leaks.)
+    erlang:trace_pattern({erlang, term_to_binary, 2}, true, []),
+    Map = huge_map(1_000_000, []),
+    rinse_and_repeat(2, Map).
+
+rinse_and_repeat(Wait, Map) ->
+    {Pid,Ref} = spawn_monitor(fun() ->
+                                      receive go -> ok end,
+                                      term_to_binary(Map, [deterministic])
+                              end),
+    erlang:trace(Pid, true, [call,arity]),
+
+    Pid ! go,
+    receive
+        {trace,Pid,call,{erlang,term_to_binary,2}} ->
+            %% Wait a short period to avoid killing term_to_binary/2 during
+            %% size calculation part.
+            receive after Wait -> ok end,
+            exit(Pid, kill);
+        Other ->
+            error(Other)
+    end,
+
+    receive
+        {'DOWN',Ref,process,Pid,Reason} ->
+            case Reason of
+                normal ->
+                    %% The encoding operation finished before the process
+                    %% was forcefully killed. Done.
+                    ok;
+                killed ->
+                    %% The process was killed in term_to_binary/2. We
+                    %% can't know whether it was killed too early, so
+                    %% we will try again with a slightly longer time.
+                    rinse_and_repeat(Wait + Wait div 2, Map)
+            end;
+        Other2 ->
+            error(Other2)
+    end.
+
+huge_map(0, Acc) ->
+    maps:from_list(Acc);
+huge_map(N, Acc) ->
+    %% Create awkward keys to slow down the sorting.
+    Key = <<N:127/big>>,
+    huge_map(N-1, [{Key,N}|Acc]).
+
+t2b_deterministic_1(0, Map) ->
+    Map;
+t2b_deterministic_1(N, Map0) ->
+    test_deterministic(Map0),
+    Map = Map0#{random_term() => N},
+    t2b_deterministic_1(N - 1, Map).
+
+test_deterministic(Map) ->
+    Bin = term_to_binary(Map, [deterministic]),
+    Map = binary_to_term(Bin),
+
+    %% Make sure that the keys for the map are ordered.
+    List0 = maps:to_list(Map),
+    List = lists:sort(fun(A, B) -> erts_internal:cmp_term(A, B) =< 0 end, List0),
+    List = decode_ext_map(Bin),
+    ok.
+
+decode_ext_map(MapBin) ->
+    MapTag = 116,
+    ListTag = 108,
+    NilTag = 106,
+
+    %% Rewrite the map to a list before decoding to preserve the order
+    %% the map elements.
+    <<131,MapTag,Size:32,Bin0/binary>> = MapBin,
+    ListBin = <<131,ListTag,(2*Size):32,Bin0/binary,NilTag>>,
+    List = binary_to_term(ListBin),
+    decode_ext_map_1(List).
+
+decode_ext_map_1([Key,Val|T]) ->
+    [{Key,Val}|decode_ext_map_1(T)];
+decode_ext_map_1([]) -> [].
+
+make_map(N) ->
+    maps:from_list([{I,I*I} || I <- lists:seq(1, N)]).
+
+random_term() ->
+    case rand:uniform(11) of
+        1 ->
+            list_to_atom(integer_to_list(36#aaaa + rand:uniform(100) * 23, 36));
+        2 ->
+            rand:uniform();
+        3 ->
+            rand:uniform(10000000);
+        4 ->
+            (1 bsl rand:uniform(128)) + rand:uniform(10000000);
+        5 ->
+            self();
+        6 ->
+            make_ref();
+        7 ->
+            random_list();
+        8 ->
+            list_to_tuple(random_list());
+        9 ->
+            [];
+        10 ->
+            make_map(33);
+        11 ->
+            make_map(rand:uniform(10))
+    end.
+
+random_list() ->
+    [random_term() || _ <- lists:seq(1, rand:uniform(10)-1)].
+
 external_size(Config) when is_list(Config) ->
     %% Build a term whose external size only fits in a big num (on 32-bit CPU).
     external_size_1(16#11111111111111117777777777777777888889999, 0, 16#FFFFFFF),
@@ -889,8 +1047,14 @@ total_memory() ->
     end.
 
 run_when_enough_resources(Fun) ->
+    DemandGb = case erlang:system_info(build_type) of
+                   Gb when Gb =:= valgrind; Gb =:= asan ->
+                       30;
+                   _ ->
+                       15
+               end,
     case {total_memory(), erlang:system_info(wordsize)} of
-        {Mem, 8} when is_integer(Mem) andalso Mem >= 15 ->
+        {Mem, 8} when is_integer(Mem) andalso Mem >= DemandGb ->
             Fun();
         {Mem, WordSize} ->
             {skipped, 
@@ -901,7 +1065,7 @@ run_when_enough_resources(Fun) ->
 
 %% OTP-4053
 bad_binary_to_term_2(Config) when is_list(Config) ->
-    {ok, N} = test_server:start_node(plopp, slave, []),
+    {ok, Peer, N} = ?CT_PEER(),
     R = rpc:call(N, erlang, binary_to_term, [<<131,111,255,255,255,0>>]),
     case R of
 	      {badrpc, {'EXIT', _}} ->
@@ -909,7 +1073,7 @@ bad_binary_to_term_2(Config) when is_list(Config) ->
 	      _Other ->
 		  ct:fail({rpcresult, R})
 	  end,
-    test_server:stop_node(N),
+    peer:stop(Peer),
     ok.
 
 %% Try bad input to binary_to_term/1.
@@ -1381,11 +1545,14 @@ test_terms(Test_Func) ->
     Test_Func("abcdef"),
     Test_Func([a, b, 1, 2]),
     Test_Func([a|b]),
+    Test_Func([make_port(), make_ref(), make_pid(), fun() -> ok end,
+               Very_Big | lists:seq(1, 75)]),
 
     Test_Func({}),
     Test_Func({1}),
     Test_Func({a, b}),
     Test_Func({a, b, c}),
+    Test_Func({make_port(), make_ref(), make_pid(), fun() -> ok end}),
     Test_Func(list_to_tuple(lists:seq(0, 255))),
     Test_Func(list_to_tuple(lists:seq(0, 256))),
 
@@ -1436,12 +1603,30 @@ test_terms(Test_Func) ->
     Test_Func(<<42:10>>),
     Test_Func(list_to_bitstring([<<5:6>>|lists:seq(0, 255)])),
 
+    %% Funs in a list.
     Test_Func(F = fun(A) -> 42*A end),
     Test_Func(lists:duplicate(32, F)),
 
+    %% External funs in a list.
     Test_Func(FF = fun binary_SUITE:all/0),
     Test_Func(lists:duplicate(32, FF)),
 
+    %% Maps.
+    SmallMap = #{a => 1, b => 2, c => 3},
+    LargeMap1 = maps:from_list([{list_to_atom(integer_to_list(36#cafe+N*N*N, 36)),N} ||
+                                   N <- lists:seq(1, 33)]),
+    LargeMap2 = maps:from_list([{list_to_atom(integer_to_list(36#dead+N, 36)),N} ||
+                                   N <- lists:seq(1, 50)]),
+    MapWithMap = LargeMap1#{SmallMap => a, LargeMap1 => LargeMap2, LargeMap2 => LargeMap1,
+                            [LargeMap1,LargeMap2] => LargeMap1,
+                            <<"abc">> => SmallMap, <<"qrs">> => LargeMap1,
+                            <<"xyz">> => LargeMap2 },
+    Test_Func(#{}),
+    Test_Func(SmallMap),
+    Test_Func(LargeMap1),
+    Test_Func(LargeMap2),
+    Test_Func(MapWithMap),
+
     ok.
 
 test_floats(Test_Func) ->
@@ -1563,29 +1748,6 @@ deep_roundtrip(T) ->
     B = term_to_binary(T),
     T = binary_to_term(B).
 
-term2bin_tuple_fallbacks(Config) when is_list(Config) ->
-    erts_debug:set_internal_state(available_internal_state, true),
-
-    term2bin_tf(fun ?MODULE:all/1),
-    term2bin_tf(<<1:1>>),
-    term2bin_tf(<<90,80:7>>),
-
-    erts_debug:set_internal_state(available_internal_state, false),
-    ok.
-
-term2bin_tf(Term) ->
-    Tuple = case Term of
-                Fun when is_function(Fun) ->
-                    {type, external} = erlang:fun_info(Fun, type),
-                    {module,M} = erlang:fun_info(Fun, module),
-                    {name,F} = erlang:fun_info(Fun, name),
-                    {M,F};
-                BS when bit_size(BS) rem 8 =/= 0 ->
-                    Bits = bit_size(BS) rem 8,
-                    {<<BS/bitstring, 0:(8-Bits)>>, Bits}
-            end,
-    Tuple = binary_to_term_stress(erts_debug:get_internal_state({term_to_binary_tuple_fallbacks,Term})).
-
 %% Test non-standard encodings never generated by term_to_binary/1
 %% but recognized by binary_to_term/1.
 
@@ -1794,18 +1956,11 @@ cmp_old_impl(Config) when is_list(Config) ->
     %% implementation in R16B. Since OTP 22 we can't talk distribution with such
     %% old nodes (< 19). The test case it kept but compares with previous major
     %% version for semantic regression test.
-    Cookie = atom_to_list(erlang:get_cookie()),
-    Rel = (integer_to_list(list_to_integer(erlang:system_info(otp_release)) - 1)
-           ++ "_latest"),
-    case test_server:is_release_available(Rel) of
-	false ->
+    Rel = integer_to_list(list_to_integer(erlang:system_info(otp_release)) - 1),
+    case ?CT_PEER([], Rel, proplists:get_value(priv_dir, Config)) of
+	not_available ->
 	    {skipped, "No OTP "++Rel++" available"};
-	true ->
-	    {ok, Node} = test_server:start_node(list_to_atom(atom_to_list(?MODULE)++"_"++Rel),
-				       peer,
-				       [{args, " -setcookie "++Cookie},
-					{erl, [{release, Rel}]}]),
-
+        {ok, Peer, Node}  ->
 	    cmp_node(Node, {erlang, list_to_binary, [list2iolist(mk_list(1))]}),
 	    cmp_node(Node, {erlang, list_to_binary, [list2iolist(mk_list(10))]}),
 	    cmp_node(Node, {erlang, list_to_binary, [list2iolist(mk_list(100))]}),
@@ -1843,7 +1998,7 @@ cmp_old_impl(Config) when is_list(Config) ->
 	    cmp_node(Node, {erlang, bitstring_to_list, [list_to_bitstring(list2bitstrlist(mk_list(1000000)))]}),
 	    cmp_node(Node, {erlang, bitstring_to_list, [list_to_bitstring(list2bitstrlist(mk_list(10000000)))]}),
 
-	    test_server:stop_node(Node),
+	    peer:stop(Peer),
 
 	    ok
     end.
