@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %% 
-%% Copyright Ericsson AB 2001-2016. All Rights Reserved.
+%% Copyright Ericsson AB 2001-2021. All Rights Reserved.
 %% 
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -45,7 +45,11 @@
          otp_5772_dist_link/1,
          otp_5772_monitor/1,
          otp_5772_dist_monitor/1,
-         otp_7946/1]).
+         otp_7946/1,
+         otp_17127_local_link_with_simultaneous_link_unlink/1,
+         otp_17127_dist_link_with_simultaneous_link_unlink/1,
+         otp_17127_local_random/1,
+         otp_17127_dist_random/1]).
 
 -export([init_per_testcase/2, end_per_testcase/2]).
 
@@ -53,26 +57,18 @@
 -export([test_proc/0]).
 
 
--define(LINK_UNDEF, 0).
--define(LINK_PID,   1).
--define(LINK_NODE,  3).
-
-
-% These are to be kept in sync with erl_monitors.h 
--define(MON_ORIGIN, 1).
--define(MON_TARGET, 3).
-
-
--record(erl_link, {type = ?LINK_UNDEF,
+-record(erl_link, {type,           % process | port | dist_process
                    pid = [],
-                   targets = []}).
+                   state,          % linked | unlinking
+                   id}).
 
 % This is to be kept in sync with erl_bif_info.c (make_monitor_list)
 
--record(erl_monitor, {type,        % MON_ORIGIN or MON_TARGET (1 or 3)
-                      ref,
+-record(erl_monitor, {type,        % process | port | time_offset | dist_process | resource | node | nodes | suspend
+                      dir,         % origin | target
+                      ref,         % Reference | []
                       pid,         % Process or nodename
-                      name = []}). % registered name or []
+                      extra = []}). % registered name, integer or, []
 
 
 suite() ->
@@ -83,7 +79,10 @@ all() ->
     [links, dist_links, monitor_nodes, process_monitors,
      dist_process_monitors, busy_dist_port_monitor,
      busy_dist_port_link, otp_5772_link, otp_5772_dist_link,
-     otp_5772_monitor, otp_5772_dist_monitor, otp_7946].
+     otp_5772_monitor, otp_5772_dist_monitor, otp_7946,
+     otp_17127_local_link_with_simultaneous_link_unlink,
+     otp_17127_dist_link_with_simultaneous_link_unlink,
+     otp_17127_local_random, otp_17127_dist_random].
 
 init_per_testcase(Func, Config) when is_atom(Func), is_list(Config) ->
     case catch erts_debug:get_internal_state(available_internal_state) of
@@ -106,18 +105,20 @@ end_per_suite(_Config) ->
 links(Config) when is_list(Config) ->
     common_link_test(node(), node()),
     true = link(self()),
-    [] = find_erl_link(self(), ?LINK_PID, self()),
+    [] = find_erl_link(self(), process, self()),
     true = unlink(self()),
     ok.
 
 %% Tests distributed links
 dist_links(Config) when is_list(Config) ->
-    [NodeName] = get_names(1, dist_link),
-    {ok, Node} = start_node(NodeName),
+    {ok, _Peer, Node} = ?CT_PEER(),
+    rpc:call(Node, erts_debug, set_internal_state, [available_internal_state, true]),
     common_link_test(node(), Node),
     TP4 = spawn(?MODULE, test_proc, []),
     TP5 = spawn(?MODULE, test_proc, []),
     TP6 = spawn(Node, ?MODULE, test_proc, []),
+    io:format("TP4=~p~nTP5=~p~nTP6=~p~n", [TP4, TP5, TP6]),
+
     true = tp_call(TP6, fun() -> link(TP4) end),
     check_link(TP4, TP6),
     true = tp_call(TP5,
@@ -173,11 +174,11 @@ common_link_test(NodeA, NodeB) ->
 
 %% Tests monitor of nodes
 monitor_nodes(Config) when is_list(Config) ->
-    [An, Bn, Cn, Dn] = get_names(4, dist_link),
-    {ok, A} = start_node(An),
-    {ok, B} = start_node(Bn),
-    C = list_to_atom(lists:concat([Cn, "@", hostname()])),
-    D = list_to_atom(lists:concat([Dn, "@", hostname()])),
+    {ok, PeerA, A} = ?CT_PEER(),
+    {ok, PeerB, B} = ?CT_PEER(),
+    [_, HostName] = string:lexemes(atom_to_list(node()), "@"),
+    C = list_to_atom(peer:random_name(?MODULE_STRING) ++ "@" ++ HostName),
+    D = list_to_atom(peer:random_name(?MODULE_STRING) ++ "@" ++ HostName),
     0 = no_of_monitor_node(self(), A),
     0 = no_of_monitor_node(self(), B),
     monitor_node(A, true),
@@ -191,6 +192,7 @@ monitor_nodes(Config) when is_list(Config) ->
     monitor_node(A, false),
     monitor_node(B, true),
     monitor_node(C, true),
+    receive after 1000 -> ok end,
     monitor_node(C, false),
     monitor_node(C, true),
     monitor_node(B, true),
@@ -200,18 +202,21 @@ monitor_nodes(Config) when is_list(Config) ->
     monitor_node(A, true),
     check_monitor_node(self(), A, 1),
     check_monitor_node(self(), B, 3),
+    ok = receive {nodedown, C} -> ok after 1000 -> timeout end,
+    %%OTP-21: monitor_node(_,false) does not trigger auto-connect anymore
+    %%        and therefore no nodedown if it fails.
+    %%ok = receive {nodedown, C} -> ok after 1000 -> timeout end,
+    ok = receive {nodedown, C} -> ok after 1000 -> timeout end,
+    ok = receive {nodedown, D} -> ok after 1000 -> timeout end,
+    ok = receive {nodedown, D} -> ok after 1000 -> timeout end,
     check_monitor_node(self(), C, 0),
     check_monitor_node(self(), D, 0),
-    receive {nodedown, C} -> ok end,
-    receive {nodedown, C} -> ok end,
-    receive {nodedown, C} -> ok end,
-    receive {nodedown, D} -> ok end,
-    receive {nodedown, D} -> ok end,
-    stop_node(A),
+
+    peer:stop(PeerA),
     receive {nodedown, A} -> ok end,
     check_monitor_node(self(), A, 0),
     check_monitor_node(self(), B, 3),
-    stop_node(B),
+    peer:stop(PeerB),
     receive {nodedown, B} -> ok end,
     receive {nodedown, B} -> ok end,
     receive {nodedown, B} -> ok end,
@@ -230,7 +235,8 @@ process_monitors(Config) when is_list(Config) ->
     common_process_monitors(node(), node()),
     Mon1 = erlang:monitor(process,self()),
     [] = find_erl_monitor(self(), Mon1),
-    [Name] = get_names(1, process_monitors),
+    Name = list_to_atom(lists:concat([?FUNCTION_NAME,
+        integer_to_list(erlang:unique_integer([positive]))])),
     true = register(Name, self()),
     Mon2 = erlang:monitor(process, Name),
     [] = find_erl_monitor(self(), Mon2),
@@ -247,8 +253,8 @@ process_monitors(Config) when is_list(Config) ->
 
 %% Tests distributed process monitors
 dist_process_monitors(Config) when is_list(Config) -> 
-    [Name] = get_names(1,dist_process_monitors),
-    {ok, Node} = start_node(Name),
+    {ok, _Peer, Node} = ?CT_PEER(),
+    rpc:call(Node, erts_debug, set_internal_state, [available_internal_state, true]),
     common_process_monitors(node(), Node),
     TP1 = spawn(Node, ?MODULE, test_proc, []),
     R1 = erlang:monitor(process, TP1),
@@ -275,7 +281,8 @@ common_process_monitors(NodeA, NodeB) ->
     run_common_process_monitors(TP1, TP2),
     TP3 = spawn(NodeA, ?MODULE, test_proc, []),
     TP4 = spawn(NodeB, ?MODULE, test_proc, []),
-    [TP4N] = get_names(1, common_process_monitors),
+    TP4N = list_to_atom(lists:concat([?FUNCTION_NAME,
+        integer_to_list(erlang:unique_integer([positive]))])),
     true = tp_call(TP4, fun () -> register(TP4N,self()) end),
     run_common_process_monitors(TP3,
                                 case node() == node(TP4) of
@@ -301,7 +308,8 @@ run_common_process_monitors(TP1, TP2) ->
     wait_until(fun () -> is_proc_dead(TP2) end),
     ok = tp_call(TP1, fun () ->
                               receive
-                                  {'DOWN',R2,process,TP2O,bye} ->
+                                  {'DOWN',R2,process,TP2O,Reason1} ->
+                                      bye = Reason1,
                                       ok
                               end
                       end),
@@ -310,7 +318,8 @@ run_common_process_monitors(TP1, TP2) ->
     R3 = tp_call(TP1, fun () -> erlang:monitor(process, TP2) end),
     ok = tp_call(TP1, fun () ->
                               receive
-                                  {'DOWN',R3,process,TP2O,noproc} ->
+                                  {'DOWN',R3,process,TP2O,Reason2} ->
+                                      noproc = Reason2,
                                       ok
                               end
                       end),
@@ -330,8 +339,7 @@ busy_dist_port_monitor(Config) when is_list(Config) ->
                  _ -> false
              end,
 
-    [An] = get_names(1, busy_dist_port_monitor),
-    {ok, A} = start_node(An),
+    {ok, PeerA, A} = ?CT_PEER(),
     TP1 = spawn(A, ?MODULE, test_proc, []),
     %% Check monitor over busy port
     M1 = suspend_on_busy_test(A,
@@ -360,7 +368,7 @@ busy_dist_port_monitor(Config) when is_list(Config) ->
                                end
                        end),
     tp_cast(TP1, fun () -> exit(normal) end),
-    stop_node(A),
+    peer:stop(PeerA),
     stop_busy_dist_port_tracer(Tracer),
     ok.
 
@@ -372,8 +380,7 @@ busy_dist_port_link(Config) when is_list(Config) ->
                  _ -> false
              end,
 
-    [An] = get_names(1, busy_dist_port_link),
-    {ok, A} = start_node(An),
+    {ok, PeerA, A} = ?CT_PEER(),
     TP1 = spawn(A, ?MODULE, test_proc, []),
     %% Check link over busy port
     suspend_on_busy_test(A,
@@ -406,7 +413,7 @@ busy_dist_port_link(Config) when is_list(Config) ->
                                end
                        end),
     tp_cast(TP1, fun () -> exit(normal) end),
-    stop_node(A),
+    peer:stop(PeerA),
     stop_busy_dist_port_tracer(Tracer),
     ok.
 
@@ -415,10 +422,9 @@ otp_5772_link(Config) when is_list(Config) ->
     otp_5772_link_test(node()).
 
 otp_5772_dist_link(Config) when is_list(Config) ->
-    [An] = get_names(1, otp_5772_dist_link),
-    {ok, A} = start_node(An),
+    {ok, PeerA, A} = ?CT_PEER(),
     otp_5772_link_test(A),
-    stop_node(A).
+    peer:stop(PeerA).
 
 otp_5772_link_test(Node) ->
     Prio = process_flag(priority, high),
@@ -447,11 +453,9 @@ otp_5772_monitor(Config) when is_list(Config) ->
     otp_5772_monitor_test(node()).
 
 otp_5772_dist_monitor(Config) when is_list(Config) ->
-    [An] = get_names(1, otp_5772_dist_monitor),
-    {ok, A} = start_node(An),
+    {ok, PeerA, A} = ?CT_PEER(),
     otp_5772_monitor_test(A),
-    stop_node(A),
-    ok.
+    peer:stop(PeerA).
 
 otp_5772_monitor_test(Node) ->
     Prio = process_flag(priority, high),
@@ -475,8 +479,7 @@ otp_5772_monitor_test(Node) ->
     ok.
 
 otp_7946(Config) when is_list(Config) ->
-    [NodeName] = get_names(1, otp_7946),
-    {ok, Node} = start_node(NodeName),
+    {ok, _Peer, Node} = ?CT_PEER(),
     Proc = rpc:call(Node, erlang, whereis, [net_kernel]),
     Mon = erlang:monitor(process, Proc),
     rpc:cast(Node, erlang, halt, []),
@@ -493,6 +496,118 @@ otp_7946(Config) when is_list(Config) ->
             Reason = noconnection
     end.
 
+otp_17127_local_link_with_simultaneous_link_unlink(Config) when is_list(Config) ->
+    otp_17127_link_with_simultaneous_link_unlink_test(node(), node()).
+
+otp_17127_dist_link_with_simultaneous_link_unlink(Config) when is_list(Config) ->
+    {ok, Peer, Node} = ?CT_PEER(),
+    Res = otp_17127_link_with_simultaneous_link_unlink_test(node(), Node),
+    peer:stop(Peer),
+    Res.
+
+otp_17127_link_with_simultaneous_link_unlink_test(NodeA, NodeB) ->
+    FunA = fun (Other) ->
+                   link(Other)
+           end,
+    FunB = fun (Other) ->
+                   link(Other),
+                   unlink(Other)
+           end,
+    otp_17127_test(NodeA, FunA, NodeB, FunB).
+
+otp_17127_local_random(Config) when is_list(Config) ->
+    otp_17127_random_test(node(), node(), 100).
+
+otp_17127_dist_random(Config) when is_list(Config) ->
+    {ok, Peer, Node} = ?CT_PEER(),
+    rpc:call(Node, erts_debug, set_internal_state, [available_internal_state, true]),
+    Res = otp_17127_random_test(node(), Node, 20),
+    peer:stop(Peer),
+    Res.
+
+otp_17127_random_test(_NodeA, _NodeB, 0) ->
+    ok;
+otp_17127_random_test(NodeA, NodeB, N) ->
+    Fun = fun (Other) ->
+                  rand_proc(Other, rand:uniform(500))
+          end,
+    otp_17127_test(NodeA, Fun, NodeB, Fun),
+    otp_17127_random_test(NodeA, NodeB, N-1).
+
+rand_proc(_Other, 0) ->
+    ok;
+rand_proc(Other, N) ->
+    case rand:uniform(3) of
+        1 -> link(Other);
+        2 -> unlink(Other);
+        3 -> erlang:yield()
+    end,
+    rand_proc(Other, N-1).
+
+otp_17127_test(NodeA, FunA, NodeB, FunB) ->
+    process_flag(priority, high),
+    {SchedA, SchedB} = case NodeA == NodeB of
+                           false ->
+                               {[], []};
+                           true ->
+                               NS = erlang:system_info(schedulers_online),
+                               process_flag(scheduler, 1),
+                               {[{scheduler, (1 rem NS) + 1}],
+                                [{scheduler, (2 rem NS) + 1}]}
+                       end,
+    A = spawn_opt(NodeA,
+                  fun () ->
+                          receive
+                              {go, Tester, Other, Later} ->
+                                  busy_wait_until(Later),
+                                  FunA(Other),
+                                  receive ping -> Other ! pong end,
+                                  receive pling -> ok end,
+                                  Tester ! {self(), done}
+                          end,
+                          receive after infinity -> ok end
+                  end, SchedA),
+    B = spawn_opt(NodeB,
+                  fun () ->
+                          receive
+                              {go, Tester, Other, Later} ->
+                                  busy_wait_until(Later),
+                                  FunB(Other),
+                                  Other ! ping,
+                                  receive pong -> Other ! pling end,
+                                  Tester ! {self(), done}
+                          end,
+                          receive after infinity -> ok end
+                  end, SchedB),
+    io:format("A = ~p~nB = ~p~n", [A, B]),
+    Later = case NodeA == NodeB of
+                true ->
+                    GoTime = (erlang:monotonic_time()
+                              + erlang:convert_time_unit(100, millisecond, native)),
+                    fun () ->
+                            erlang:monotonic_time() >= GoTime
+                    end;
+                false ->
+                    GoTime = (os:system_time(nanosecond)
+                              + erlang:convert_time_unit(500, millisecond, nanosecond)),
+                    fun () ->
+                            os:system_time(nanosecond) >= GoTime
+                    end
+            end,
+    erlang:yield(),
+    A ! {go, self(), B, Later},
+    B ! {go, self(), A, Later},
+    receive {A, done} -> ok end,
+    receive {B, done} -> ok end,
+    try
+        check_consistent_link_state(A, B),
+        true = rpc:call(node(A), erlang, is_process_alive, [A]),
+        true = rpc:call(node(B), erlang, is_process_alive, [B])
+    after
+        exit(A, kill),
+        exit(B, kill)
+    end.
+
 %%
 %% -- Internal utils --------------------------------------------------------
 %%
@@ -533,7 +648,7 @@ freeze_node(Node, MS) ->
                fun () ->
                        erts_debug:set_internal_state(available_internal_state,
                                                      true),
-                       dport_send(Freezer, DoingIt),
+                       dctrl_dop_send(Freezer, DoingIt),
                        receive after Own -> ok end,
                        erts_debug:set_internal_state(block, MS+Own)
                end),
@@ -544,20 +659,22 @@ make_busy(Node, Time) when is_integer(Time) ->
     Own = 500,
     freeze_node(Node, Time+Own), 
     Data = busy_data(),
+    DCtrl = dctrl(Node),
     %% first make port busy
     Pid = spawn_link(fun () ->
                              forever(fun () ->
-                                             dport_reg_send(Node,
-                                                            '__noone__',
-                                                            Data)
+                                             dctrl_dop_reg_send(Node,
+                                                                '__noone__',
+                                                                Data)
                                      end)
                      end),
     receive after Own -> ok end,
     wait_until(fun () ->
-                       case process_info(Pid, status) of
-                           {status, suspended} -> true;
-                           _ -> false
-                       end
+                  case {DCtrl, process_info(Pid, status)} of
+                      {DPrt, {status, suspended}} when is_port(DPrt) -> true;
+                      {DPid, {status, waiting}} when is_pid(DPid) -> true;
+                      _ -> false
+                  end
                end),
     %% then dist entry
     make_busy(Node, [nosuspend], Data),
@@ -669,6 +786,13 @@ wait_until(Fun) ->
             end
     end.
 
+busy_wait_until(Fun) ->
+    case Fun() of
+        true -> ok;
+        _ ->
+            busy_wait_until(Fun)
+    end.
+
 forever(Fun) ->
     Fun(),
     forever(Fun).
@@ -693,22 +817,10 @@ test_proc() ->
     end,
     test_proc().
 
-expand_link_list([#erl_link{type = ?LINK_NODE, targets = N} = Rec | T]) ->
-    lists:duplicate(N,Rec#erl_link{targets = []}) ++ expand_link_list(T);
-expand_link_list([#erl_link{targets = [#erl_link{pid = Pid}]} = Rec | T]) ->
-    [Rec#erl_link{targets = [Pid]} | expand_link_list(T)];
-expand_link_list([#erl_link{targets = [#erl_link{pid = Pid}|TT]} = Rec | T]) ->
-    [ Rec#erl_link{targets = [Pid]} | expand_link_list( 
-                                        [Rec#erl_link{targets = TT} | T])]; 
-expand_link_list([#erl_link{targets = []} = Rec | T]) ->
-    [Rec | expand_link_list(T)];
-expand_link_list([]) ->
-    [].
-
 get_local_link_list(Obj) ->
     case catch erts_debug:get_internal_state({link_list, Obj}) of
         LL when is_list(LL) ->
-            expand_link_list(LL);
+            LL;
         _ ->
             []
     end.
@@ -717,7 +829,7 @@ get_remote_link_list(Node, Obj) ->
     case catch rpc:call(Node, erts_debug, get_internal_state,
                         [{link_list, Obj}]) of
         LL when is_list(LL) ->
-            expand_link_list(LL);
+            LL;
         _ ->
             []
     end.
@@ -771,82 +883,125 @@ get_monitor_list(undefined) ->
 
 
 find_erl_monitor(Pid, Ref) when is_reference(Ref) ->
+    MonitorList = get_monitor_list(Pid),
+    io:format("~p MonitorList: ~p~n", [Pid, MonitorList]),
     lists:foldl(fun (#erl_monitor{ref = R} = EL, Acc) when R == Ref ->
                         [EL|Acc];
                     (_, Acc) ->
                         Acc
                 end,
                 [],
-                get_monitor_list(Pid)).
-
-% find_erl_link(Obj, Ref) when is_reference(Ref) -> 
-%     lists:foldl(fun (#erl_link{ref = R} = EL, Acc) when R == Ref ->
-% 			      [EL|Acc];
-% 			  (_, Acc) ->
-% 			      Acc
-% 		      end,
-% 		      [],
-% 		      get_link_list(Obj)).
-
-find_erl_link(Obj, Type, [Item, Data]) when is_pid(Item);
-                                            is_port(Item);
-                                            is_atom(Item) -> 
-    lists:foldl(fun (#erl_link{type = T, pid = I, targets = D} = EL,
-                     Acc) when T == Type, I == Item ->
-                        case Data of
-                            D ->
-                                [EL|Acc];
-                            [] ->
-                                [EL|Acc];
-                            _ ->
-                                Acc
-                        end;
+                MonitorList);
+find_erl_monitor(Pid, Item) ->
+    MonitorList = get_monitor_list(Pid),
+    io:format("~p MonitorList: ~p~n", [Pid, MonitorList]),
+    lists:foldl(fun (#erl_monitor{pid = I} = EL, Acc) when I == Item ->
+                        [EL|Acc];
                     (_, Acc) ->
                         Acc
                 end,
                 [],
-                get_link_list(Obj));
-find_erl_link(Obj, Type, Item) when is_pid(Item); is_port(Item); is_atom(Item) ->
-    find_erl_link(Obj, Type, [Item, []]). 
+                MonitorList).
 
 
+find_erl_link(Obj, Type, Item) when is_pid(Item); is_port(Item) -> 
+    LinkList = get_link_list(Obj),
+    io:format("~p LinkList: ~p~n", [Obj, LinkList]),
+    lists:foldl(fun (#erl_link{type = T, pid = I, state = linked} = EL,
+                     Acc) when T == Type, I == Item ->
+                        [EL|Acc];
+                    ({erl_link, T, P, I},
+                     Acc) when T == Type, P == Item ->
+                        %% Old emulator without state (linked if record exists)...
+                        [#erl_link{type = T, pid = P, id = I, state = linked}|Acc];
+                    (_, Acc) ->
+                        Acc
+                end,
+                [],
+                LinkList);
+find_erl_link(Obj, Type, Id) when is_integer(Id) -> 
+    %% Find by Id
+    LinkList = get_link_list(Obj),
+    io:format("~p LinkList: ~p~n", [Obj, LinkList]),
+    lists:foldl(fun (#erl_link{type = T, id = I, state = linked} = EL,
+                     Acc) when T == Type, I == Id ->
+                        [EL|Acc];
+                    ({erl_link, T, P, I},
+                     Acc) when T == Type, I == Id ->
+                        %% Old emulator without state (linked if record exists)...
+                        [#erl_link{type = T, pid = P, id = I, state = linked}|Acc];
+                    (_, Acc) ->
+                        Acc
+                end,
+                [],
+                LinkList).
 
-check_link(A, B) ->
-    [#erl_link{type = ?LINK_PID,
-               pid = B,
-               targets = []}] = find_erl_link(A, ?LINK_PID, B),
-    [#erl_link{type = ?LINK_PID,
-               pid = A,
-               targets = []}] = find_erl_link(B, ?LINK_PID, A),
+
+get_link_type(A, B) when is_port(A);
+                         is_port(B) ->
+    port;
+get_link_type(A, B) when is_pid(A),
+                         is_pid(B) ->
     case node(A) == node(B) of
-        false ->
-            [#erl_link{type = ?LINK_PID,
-                       pid = A,
-                       targets = [B]}] = find_erl_link({node(A),
-                                                        node(B)},
-                                                       ?LINK_PID,
-                                                       [A, [B]]),
-            [#erl_link{type = ?LINK_PID,
-                       pid = B,
-                       targets = [A]}] = find_erl_link({node(B),
-                                                        node(A)},
-                                                       ?LINK_PID,
-                                                       [B, [A]]);
         true ->
-            [] = find_erl_link({node(A), node(B)},
-                               ?LINK_PID,
-                               [A, [B]]),
-            [] = find_erl_link({node(B), node(A)},
-                               ?LINK_PID,
-                               [B, [A]])
-    end,
+            process;
+        false ->
+            dist_process
+    end.
+
+check_consistent_link_state(A, B) ->
+    %% Both processes should agree on whether
+    %% they are linked or not...
+    LinkType = get_link_type(A, B),
+    case find_erl_link(A, LinkType, B) of
+        [] ->
+            check_unlink(A, B),
+            io:format("~p and ~p are not linked~n", [A, B]);
+        _ ->
+            check_link(A, B),
+            io:format("~p and ~p are linked~n", [A, B])
+    end.
+
+check_link(A, B) when node(A) == node(B) ->
+    LinkType = get_link_type(A, B),
+    [#erl_link{type = LinkType,
+               pid = B}] = find_erl_link(A, LinkType, B),
+    [#erl_link{type = LinkType,
+               pid = A}] = find_erl_link(B, LinkType, A),
+    [] = find_erl_link({node(A), node(B)},
+                       LinkType,
+                       A),
+    [] = find_erl_link({node(B), node(A)},
+                       LinkType,
+                       B),
+    ok;
+check_link(A, B) ->
+    [#erl_link{type = dist_process,
+               pid = B,
+               id = IdA}] = find_erl_link(A, dist_process, B),
+    [#erl_link{type = dist_process,
+               pid = A,
+               id = IdA}] = find_erl_link({node(A),
+                                           node(B)},
+                                          dist_process,
+                                          IdA),
+    [#erl_link{type = dist_process,
+               pid = A,
+               id = IdB}] = find_erl_link(B, dist_process, A),
+    [#erl_link{type = dist_process,
+               pid = B,
+               id = IdB}] = find_erl_link({node(B),
+                                           node(A)},
+                                          dist_process,
+                                          IdB),
     ok.
 
 check_unlink(A, B) ->
-    [] = find_erl_link(A, ?LINK_PID, B),
-    [] = find_erl_link(B, ?LINK_PID, A),
-    [] = find_erl_link({node(A), node(B)}, ?LINK_PID, [A, [B]]),
-    [] = find_erl_link({node(B), node(A)}, ?LINK_PID, [B, [A]]),
+    LinkType = get_link_type(A, B),
+    [] = find_erl_link(A, LinkType, B),
+    [] = find_erl_link(B, LinkType, A),
+    [] = find_erl_link({node(A), node(B)}, dist_process, A),
+    [] = find_erl_link({node(B), node(A)}, dist_process, B),
     ok.
 
 check_process_monitor(From, {Name, Node}, Ref) when is_pid(From),
@@ -859,22 +1014,26 @@ check_process_monitor(From, {Name, Node}, Ref) when is_pid(From),
                                                     is_atom(Node),
                                                     is_reference(Ref) ->
     MonitoredPid = rpc:call(Node, erlang, whereis, [Name]),
-    [#erl_monitor{type = ?MON_ORIGIN,
+    [#erl_monitor{type = dist_process,
+                  dir = origin,
                   ref = Ref,
                   pid = Node,
-                  name = Name}] = find_erl_monitor(From, Ref),
-    [#erl_monitor{type = ?MON_TARGET,
+                  extra = Name}] = find_erl_monitor(From, Ref),
+    [#erl_monitor{type = dist_process,
+                  dir = target,
                   ref = Ref,
                   pid = From,
-                  name = Name}] = 	find_erl_monitor({node(From), Node}, Ref),
-    [#erl_monitor{type = ?MON_ORIGIN,
+                  extra = Name}] = find_erl_monitor({node(From), Node}, Ref),
+    [#erl_monitor{type = dist_process,
+                  dir = origin,
                   ref = Ref,
                   pid = MonitoredPid,
-                  name = Name}] = find_erl_monitor({Node, node(From)}, Ref),
-    [#erl_monitor{type = ?MON_TARGET,
+                  extra = Name}] = find_erl_monitor({Node, node(From)}, Ref),
+    [#erl_monitor{type = dist_process,
+                  dir = target,
                   ref = Ref,
                   pid = From,
-                  name = Name}] = find_erl_monitor(MonitoredPid, Ref),
+                  extra = Name}] = find_erl_monitor(MonitoredPid, Ref),
     ok;
 check_process_monitor(From, Name, Ref) when is_pid(From),
                                             is_atom(Name),
@@ -882,27 +1041,36 @@ check_process_monitor(From, Name, Ref) when is_pid(From),
                                             is_reference(Ref) ->
     MonitoredPid = rpc:call(node(From), erlang, whereis, [Name]),
 
-    [#erl_monitor{type = ?MON_ORIGIN,
+    [#erl_monitor{type = process,
+                  dir = origin,
                   ref = Ref,
                   pid = MonitoredPid,
-                  name = Name}] = find_erl_monitor(From, Ref),
+                  extra = Name}] = find_erl_monitor(From, Ref),
 
 
-    [#erl_monitor{type = ?MON_TARGET,
+    [#erl_monitor{type = process,
+                  dir = target,
                   ref = Ref,
                   pid = From,
-                  name = Name}] = find_erl_monitor(MonitoredPid,Ref),
+                  extra = Name}] = find_erl_monitor(MonitoredPid,Ref),
     ok;
 check_process_monitor(From, To, Ref) when is_pid(From),
                                           is_pid(To),
                                           is_reference(Ref) ->
-    OriMon = [#erl_monitor{type = ?MON_ORIGIN,
+    MonType = case node(From) == node(To) of
+                  true -> process;
+                  false -> dist_process
+              end,
+
+    OriMon = [#erl_monitor{type = MonType,
+                           dir = origin,
                            ref = Ref,
                            pid = To}],
 
     OriMon = find_erl_monitor(From, Ref), 
 
-    TargMon = [#erl_monitor{type = ?MON_TARGET,
+    TargMon = [#erl_monitor{type = MonType,
+                            dir = target,
                             ref = Ref,
                             pid = From}],
     TargMon = find_erl_monitor(To, Ref),
@@ -910,7 +1078,11 @@ check_process_monitor(From, To, Ref) when is_pid(From),
 
     case node(From) == node(To) of
         false ->
-            TargMon = find_erl_monitor({node(From), node(To)}, Ref),
+            DistTargMon = [#erl_monitor{type = dist_process,
+                                        dir = target,
+                                        ref = Ref,
+                                        pid = From}],
+            DistTargMon = find_erl_monitor({node(From), node(To)}, Ref),
             OriMon = find_erl_monitor({node(To), node(From)}, Ref);
         true ->
             [] = find_erl_monitor({node(From), node(From)}, Ref)
@@ -981,54 +1153,29 @@ check_process_demonitor(From, To, Ref) when is_pid(From),
     ok.
 
 no_of_monitor_node(From, Node) when is_pid(From), is_atom(Node) ->
-    length(find_erl_link(From, ?LINK_NODE, Node)).
+    case find_erl_monitor(From, Node) of
+        [] -> 0;
+        [#erl_monitor{type = node,
+                      dir = origin,
+                      pid = Node,
+                      extra = N}] -> N
+    end.
 
+check_monitor_node(From, Node, 0) when is_pid(From),
+                                        is_atom(Node) ->
+    [] = find_erl_monitor(From, Node),
+    [] = find_erl_monitor({node(From), Node}, From);
 check_monitor_node(From, Node, No) when is_pid(From),
                                         is_atom(Node),
                                         is_integer(No),
-                                        No >= 0 ->
-    LL = lists:duplicate(No, #erl_link{type = ?LINK_NODE, pid = Node}),
-    DLL = lists:duplicate(No, #erl_link{type = ?LINK_NODE, pid = From}),
-    LL = find_erl_link(From, ?LINK_NODE, Node),
-    DLL = find_erl_link({node(From), Node}, ?LINK_NODE, From),
-    ok.
-
-
-
-hostname() ->
-    from($@, atom_to_list(node())).
-
-from(H, [H | T]) -> T;
-from(H, [_ | T]) -> from(H, T);
-from(_H, []) -> [].
-
-get_names(N, T) when is_atom(T) ->
-    get_names(N, T, []).
-get_names(0, _, Acc) ->
-    Acc;
-get_names(N, T, Acc) ->
-    get_names(N-1, T, [list_to_atom(atom_to_list(?MODULE)
-                                    ++ "-"
-                                    ++ atom_to_list(T)
-                                    ++ "-"
-                                    ++ integer_to_list(erlang:system_time(second))
-                                    ++ "-"
-                                    ++ integer_to_list(erlang:unique_integer([positive]))) | Acc]).
-
-start_node(Name) ->
-    start_node(Name, "").
-
-start_node(Name, Args) ->
-    Pa = filename:dirname(code:which(?MODULE)),
-    Res = test_server:start_node(Name, slave, [{args,  Args ++ " -pa " ++ Pa}]),
-    {ok, Node} = Res,
-    rpc:call(Node, erts_debug, set_internal_state,
-             [available_internal_state, true]),
-    Res.
-
-
-stop_node(Node) ->
-    test_server:stop_node(Node).
+                                        No > 0 ->
+    [#erl_monitor{type = node,
+                  dir = origin,
+                  pid = Node,
+                  extra = No}] = find_erl_monitor(From, Node),
+    [#erl_monitor{type = node,
+                  dir = target,
+                  pid = From}] = find_erl_monitor({node(From), Node}, From).
 
 -define(COOKIE, '').
 -define(DOP_LINK,		1).
@@ -1048,47 +1195,50 @@ stop_node(Node) ->
 -define(DOP_DEMONITOR_P,	20).
 -define(DOP_MONITOR_P_EXIT,	21).
 
-dport_send(To, Msg) ->
-    Node = node(To),
-    DPrt = case dport(Node) of
-               undefined ->
-                   pong = net_adm:ping(Node),
-                   dport(Node);
-               Prt ->
-                   Prt
-           end,
-    port_command(DPrt, [dmsg_hdr(),
-                        dmsg_ext({?DOP_SEND,
-                                  ?COOKIE,
-                                  To}),
-                        dmsg_ext(Msg)]).
+ensure_dctrl(Node) ->
+    case dctrl(Node) of
+        undefined ->
+            pong = net_adm:ping(Node),
+            dctrl(Node);
+        DCtrl ->
+            DCtrl
+    end.
 
-dport_reg_send(Node, Name, Msg) ->
-    DPrt = case dport(Node) of
-               undefined ->
-                   pong = net_adm:ping(Node),
-                   dport(Node);
-               Prt ->
-                   Prt
-           end,
-    port_command(DPrt, [dmsg_hdr(),
-                        dmsg_ext({?DOP_REG_SEND,
-                                  self(),
-                                  ?COOKIE,
-                                  Name}),
-                        dmsg_ext(Msg)]).
+dctrl_send(DPrt, Data) when is_port(DPrt) ->
+    port_command(DPrt, Data);
+dctrl_send(DPid, Data) when is_pid(DPid) ->
+    Ref = make_ref(),
+    DPid ! {send, self(), Ref, Data},
+    receive {Ref, Res} -> Res end.
 
-dport(Node) when is_atom(Node) ->
+dctrl_dop_send(To, Msg) ->
+    dctrl_send(ensure_dctrl(node(To)),
+               [dmsg_hdr(),
+                dmsg_ext({?DOP_SEND,
+                          ?COOKIE,
+                          To}),
+                dmsg_ext(Msg)]).
+
+dctrl_dop_reg_send(Node, Name, Msg) ->
+    dctrl_send(ensure_dctrl(Node),
+               [dmsg_hdr(),
+                dmsg_ext({?DOP_REG_SEND,
+                          self(),
+                          ?COOKIE,
+                          Name}),
+                dmsg_ext(Msg)]).
+
+dctrl(Node) when is_atom(Node) ->
     case catch erts_debug:get_internal_state(available_internal_state) of
         true -> true;
         _ -> erts_debug:set_internal_state(available_internal_state, true)
     end,
-    erts_debug:get_internal_state({dist_port, Node}).
+    erts_debug:get_internal_state({dist_ctrl, Node}).
 
 dmsg_hdr() ->
     [131, % Version Magic
      $D,  % Dist header
-     0].  % No atom cache referenses
+     0].  % No atom cache references
 
 dmsg_ext(Term) ->	
     <<131, Res/binary>> = term_to_binary(Term),
