@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %% 
-%% Copyright Ericsson AB 1997-2020. All Rights Reserved.
+%% Copyright Ericsson AB 1997-2022. All Rights Reserved.
 %% 
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -145,6 +145,12 @@ pending({badarg,[{erlang,Bif,BifArgs,Loc1},
   when is_atom(Bif), is_list(BifArgs), length(Args) =:= Arity,
        is_list(Loc1), is_list(Loc2) ->
     ok;
+pending({badarith,[{erlang,Bif,BifArgs,Loc1},
+                   {?MODULE,Func,Arity,Loc2}|_]},
+        Func, Args, _Code)
+  when is_atom(Bif), is_list(BifArgs), length(Args) =:= Arity,
+       is_list(Loc1), is_list(Loc2) ->
+    ok;
 pending({undef,[{non_existing_module,foo,[],Loc}|_]}, _, _, _)
   when is_list(Loc) ->
     ok;
@@ -157,7 +163,7 @@ pending({Code,[{?MODULE,Func,Arity,Loc}|_]}, Func, Args, Code)
 pending(Reason, _Function, _Args, _Code) ->
     ct:fail({bad_exit_reason,Reason}).
 
-%% Test that doing arithmetics on [] gives a badarith EXIT and not a crash.
+%% Test that doing arithmetic on [] gives a badarith EXIT and not a crash.
 
 nil_arith(Config) when is_list(Config) ->
     ba_plus_minus_times([], []),
@@ -323,6 +329,10 @@ stacktrace(Conf) when is_list(Conf) ->
         stacktrace_1({value,V}, error, {value,V}),
     {caught2,{throw,V},[{?MODULE,foo,1,_}|_]} =
         stacktrace_1({value,V}, error, {throw,V}),
+    {caught2,{error,V},[{?MODULE,foo,[a],_}|_]} =
+        stacktrace_1({value,V}, error, {error,{V,[a]}}),
+    {caught2,{error,V},[{?MODULE,foo,1,_}|_]} =
+        stacktrace_1({value,V}, error, {error,{V,none}}),
 
     try
         stacktrace_2()
@@ -442,7 +452,9 @@ foo({'add',{A,B}}) ->
     my_add(A, B);
 foo({'abs',X}) ->
     my_abs(X);
-foo({error,Error}) -> 
+foo({error,{Error, Args}}) ->
+    erlang:error(Error, Args);
+foo({error,Error}) ->
     erlang:error(Error);
 foo({throw,Throw}) ->
     erlang:throw(Throw);
@@ -674,7 +686,8 @@ do_error_3(Reason, Args, Options) ->
 
 error_info(_Config) ->
     DeadProcess = dead_process(),
-    NewAtom = non_existing_atom(),
+    {NewAtom,NewAtomExt} = non_existing_atom(),
+    Eons = 1 bsl 50,
 
     %% We'll need an incorrect memory type for erlang:memory/1. We want to test an
     %% incorrect atom if our own allocators are enabled, but if they are disabled,
@@ -686,7 +699,7 @@ error_info(_Config) ->
                     end,
 
     %% Pick up external pid and port.
-    {ok, ExternalNode} = test_server:start_node(?FUNCTION_NAME, slave, []),
+    {ok, Peer, ExternalNode} = ?CT_PEER(),
     ExternalPid = rpc:call(ExternalNode, erlang, whereis, [code_server]),
     ExternalPort = hd(rpc:call(ExternalNode, erlang, ports, [])),
 
@@ -737,7 +750,9 @@ error_info(_Config) ->
          {binary_to_existing_atom, [abc, latin1]},
          {binary_to_existing_atom, [<<128,128,255>>,utf8]},
          {binary_to_existing_atom, [list_to_binary(NewAtom), latin1]},
+         {binary_to_existing_atom, [list_to_binary(NewAtom), utf8]},
          {binary_to_existing_atom, [list_to_binary(NewAtom), utf42]},
+         {binary_to_existing_atom, [[<<"abc">>], utf8]},
          {binary_to_existing_atom, [<<0:512/unit:8>>, latin1]},
 
          {binary_to_float, [abc]},
@@ -758,9 +773,14 @@ error_info(_Config) ->
          {binary_to_list, [<<1,2,3>>, 4, 5]},
 
          {binary_to_term, [abc]},
-         {binary_to_term, [<<"abc">>]},
+         {binary_to_term, [<<"bad">>]},
+
          {binary_to_term, [term_to_binary(a), abc]},
          {binary_to_term, [<<"bad">>, abc]},
+         {binary_to_term, [<<"bad">>, [bad]]},
+         {binary_to_term, [<<"bad">>, []]},
+         {binary_to_term, [<<"bad">>, [safe]]},
+         {binary_to_term, [NewAtomExt, [safe]]},
 
          {bit_size, [abc]},
          {bitstring_to_list, [abc]},
@@ -879,6 +899,7 @@ error_info(_Config) ->
 
          {gather_gc_info_result, 1},            %Internal BIF.
 
+         {get_cookie, [{not_node}]},
          {get_keys, [value], [no_fail]},
          {get_module_info, 1},
          {get_module_info, 2},
@@ -1017,6 +1038,7 @@ error_info(_Config) ->
          {open_port, [{bad,name}, []]},
          {open_port, [{spawn, "no_command"}, bad_option_list]},
          {open_port, [{spawn, "no_command"}, [xyz]]},
+         {open_port, [{spawn, "no_command"}, [{args,[]}]],[{1,".*spawn_executable.*"}]},
 
          {port_call, 2},                        %Internal BIF.
 
@@ -1113,7 +1135,13 @@ error_info(_Config) ->
          {send, [[bad,dest], message]},
          {send, [[bad,dest], message, bad]},
 
+         {send_after, [Eons, self(), message]},
+         {send_after, [Eons, {bad,dest}, message]},
          {send_after, [bad_time, {bad,dest}, message]},
+         {send_after, [20, ExternalPid, message]},
+
+         {send_after, [Eons, self(), message, bad_options]},
+         {send_after, [Eons, {bad,dest}, message, bad_options]},
          {send_after, [bad_time, {bad,dest}, message, bad_options]},
          {send_after, [20, self(), message, [bad]]},
          {send_after, [20, ExternalPid, message, []]},
@@ -1128,7 +1156,9 @@ error_info(_Config) ->
          {seq_trace_info, 1},
          {seq_trace_print, 1},
          {seq_trace_print, 2},
-         {set_cookie, 2},
+         {set_cookie, [{not_cookie}]},
+         {set_cookie, [{not_node}, {not_cookie}]},
+         {set_cookie, [nonode@nohost, {not_cookie}]},
          {set_cpu_topology, 1},
 
          {setelement, [a, b, c]},
@@ -1192,7 +1222,12 @@ error_info(_Config) ->
          {split_binary, [a, -1]},
          {split_binary, [<<>>, 1]},
 
+         {start_timer, [Eons, self(), message]},
+         {start_timer, [Eons, {bad,dest}, message]},
          {start_timer, [bad_time, {bad,dest}, message]},
+
+         {start_timer, [Eons, self(), message, []]},
+         {start_timer, [Eons, {bad,dest}, message, [bad]]},
          {start_timer, [bad_time, {bad,dest}, message, bad_options]},
          {start_timer, [20, self(), message, [bad]]},
          {start_timer, [20, ExternalPid, message, []]},
@@ -1279,7 +1314,7 @@ error_info(_Config) ->
     try
         do_error_info(L)
     after
-        test_server:stop_node(ExternalNode)
+        peer:stop(Peer)
     end.
 
 dead_process() ->
@@ -1299,7 +1334,9 @@ non_existing_atom(Atom) ->
             non_existing_atom([Char|Atom])
     catch
         error:badarg ->
-            Atom
+            AtomBin = list_to_binary(Atom),
+            AtomExt = <<131,100,(byte_size(AtomBin)):16,AtomBin/binary>>,
+            {Atom,AtomExt}
     end.
 
 do_error_info(L0) ->
@@ -1406,11 +1443,23 @@ line_numbers(Config) when is_list(Config) ->
                [{file,"fake_file.erl"},{line,3}]},
               {?MODULE,line_numbers,1,_}|_]}} =
     (catch line1(bad_tag, 0)),
-    {'EXIT',{badarith,
-             [{?MODULE,line1,2,
-               [{file,"fake_file.erl"},{line,5}]},
-              {?MODULE,line_numbers,1,_}|_]}} =
-    (catch line1(a, not_an_integer)),
+
+    %% The stacktrace for operators such a '+' can vary depending on
+    %% whether the JIT is used or not.
+    case catch line1(a, not_an_integer) of
+        {'EXIT',{badarith,
+                 [{erlang,'+',[not_an_integer,1],_},
+                  {?MODULE,line1,2,
+                   [{file,"fake_file.erl"},{line,5}]},
+                  {?MODULE,line_numbers,1,_}|_]}} ->
+            ok;
+        {'EXIT',{badarith,
+                 [{?MODULE,line1,2,
+                   [{file,"fake_file.erl"},{line,5}]},
+                  {?MODULE,line_numbers,1,_}|_]}} ->
+            ok
+    end,
+
     {'EXIT',{{badmatch,{ok,1}},
              [{?MODULE,line1,2,
                [{file,"fake_file.erl"},{line,7}]},
@@ -1443,7 +1492,7 @@ line_numbers(Config) when is_list(Config) ->
     <<0,0>> = build_binary1(16),
     {'EXIT',{badarg,
              [{?MODULE,build_binary1,1,
-               [{file,"bit_syntax.erl"},{line,72503}]},
+               [{file,"bit_syntax.erl"},{line,72503},{error_info,_}]},
               {?MODULE,line_numbers,1,
                [{file,ModFile},{line,_}]}|_]}} =
     (catch build_binary1(bad_size)),
@@ -1451,21 +1500,20 @@ line_numbers(Config) when is_list(Config) ->
     <<7,1,2,3>> = build_binary2(8, <<1,2,3>>),
     {'EXIT',{badarg,
              [{?MODULE,build_binary2,2,
-               [{file,"bit_syntax.erl"},{line,72507}]},
+               [{file,"bit_syntax.erl"},{line,72507},{error_info,_}]},
               {?MODULE,line_numbers,1,
                [{file,ModFile},{line,_}]}|_]}} =
     (catch build_binary2(bad_size, <<>>)),
     {'EXIT',{badarg,
-             [{erlang,bit_size,[bad_binary],[{error_info,_}]},
-              {?MODULE,build_binary2,2,
-               [{file,"bit_syntax.erl"},{line,72507}]},
+             [{?MODULE,build_binary2,2,
+               [{file,"bit_syntax.erl"},{line,72507},{error_info,_}]},
               {?MODULE,line_numbers,1,
                [{file,ModFile},{line,_}]}|_]}} =
     (catch build_binary2(8, bad_binary)),
 
     <<"abc",357:16>> = build_binary3(<<"abc">>),
     {'EXIT',{badarg,[{?MODULE,build_binary3,1,
-                      [{file,"bit_syntax.erl"},{line,72511}]},
+                      [{file,"bit_syntax.erl"},{line,72511},{error_info,_}]},
                      {?MODULE,line_numbers,1,
                       [{file,ModFile},{line,_}]}|_]}} =
     (catch build_binary3(no_binary)),
@@ -1494,14 +1542,29 @@ line_numbers(Config) when is_list(Config) ->
               {?MODULE,line_numbers,1,_}|_]}} =
     (catch applied_bif_2()),
 
-    {'EXIT',{badarith,
-             [{?MODULE,increment1,1,[{file,"increment.erl"},{line,45}]},
-              {?MODULE,line_numbers,1,_}|_]}} =
-        (catch increment1(x)),
-    {'EXIT',{badarith,
-             [{?MODULE,increment2,1,[{file,"increment.erl"},{line,48}]},
-              {?MODULE,line_numbers,1,_}|_]}} =
-        (catch increment2(x)),
+    case catch increment1(x) of
+        {'EXIT',{badarith,
+                 [{erlang,'+',[x,1],_},
+                  {?MODULE,increment1,1,[{file,"increment.erl"},{line,45}]},
+                  {?MODULE,line_numbers,1,_}|_]}} ->
+            ok;
+        {'EXIT',{badarith,
+                 [{?MODULE,increment1,1,[{file,"increment.erl"},{line,45}]},
+                  {?MODULE,line_numbers,1,_}|_]}} ->
+            ok
+    end,
+
+    case catch increment2(x) of
+        {'EXIT',{badarith,
+                 [{erlang,'+',[x,1],_},
+                  {?MODULE,increment2,1,[{file,"increment.erl"},{line,48}]},
+                  {?MODULE,line_numbers,1,_}|_]}} ->
+            ok;
+        {'EXIT',{badarith,
+                 [{?MODULE,increment2,1,[{file,"increment.erl"},{line,48}]},
+                  {?MODULE,line_numbers,1,_}|_]}} ->
+            ok
+    end,
 
     {'EXIT',{{badmap,not_a_map},
              [{?MODULE,update_map,1,[{file,"map.erl"},{line,3}]}|_]}} =
@@ -1519,6 +1582,28 @@ line_numbers(Config) when is_list(Config) ->
               {?MODULE,test_tl,1,[{file,"list_bifs.erl"},{line,102}]}|_]}} =
         (catch test_tl(y)),
 
+    %% This line number is too large to be represented and will be silently
+    %% ignored by the emulator.
+    {'EXIT',{crash,
+             [{?MODULE,crash_huge_line,1,[]},
+              {?MODULE,line_numbers,1,_}|_]}} =
+        (catch crash_huge_line(gurka)),
+
+    {'EXIT',{{badrecord,[1,2,3]},
+             [{?MODULE,bad_record,1,[{file,"bad_records.erl"},{line,4}]}|_]}} =
+        catch bad_record([1,2,3]),
+
+    %% GH-5960: When an instruction raised an exception at the very end of the
+    %% instruction (e.g. badmatch), and was directly followed by a line
+    %% instruction, the exception was raised for the wrong line.
+    ok = ambiguous_line(0),
+    {'EXIT',{{badmatch,_},
+        [{?MODULE,ambiguous_line,1,[{file,"ambiguous_line.erl"},
+                                    {line,3}]}|_]}} = catch ambiguous_line(1),
+    {'EXIT',{{badmatch,_},
+        [{?MODULE,ambiguous_line,1,[{file,"ambiguous_line.erl"},
+                                    {line,4}]}|_]}} = catch ambiguous_line(2),
+
     ok.
 
 id(I) -> I.
@@ -1641,3 +1726,19 @@ update_map(M0) ->                               %Line 2
 test_hd(X) -> foo(), hd(X).                     %Line 101
 test_tl(X) -> foo(), tl(X).                     %Line 102
 foo() -> id(100).
+
+-file("huge_lines.erl", 100000000).             %Line 100000000
+
+crash_huge_line(_) ->                           %Line 100000002
+    erlang:error(crash).                        %Line 100000003
+
+-file("bad_records.erl", 1).
+-record(foobar, {a,b,c,d}).                     %Line 2.
+bad_record(R) ->                                %Line 3.
+    R#foobar.a.                                 %Line 4.
+
+-file("ambiguous_line.erl", 1).
+ambiguous_line(A) ->                            %Line 2.
+    true = A =/= 1,                             %Line 3.
+    true = A =/= 2,                             %Line 4.
+    ok.
