@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 1997-2018. All Rights Reserved.
+%% Copyright Ericsson AB 1997-2021. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -38,7 +38,8 @@
 	 system_monitor_long_gc_1/1, system_monitor_long_gc_2/1, 
 	 system_monitor_large_heap_1/1, system_monitor_large_heap_2/1,
 	 system_monitor_long_schedule/1,
-	 bad_flag/1, trace_delivered/1, trap_exit_self_receive/1]).
+	 bad_flag/1, trace_delivered/1, trap_exit_self_receive/1,
+         trace_info_badarg/1, erl_704/1, ms_excessive_nesting/1]).
 
 -include_lib("common_test/include/ct.hrl").
 
@@ -62,7 +63,8 @@ all() ->
      system_monitor_long_gc_2, system_monitor_large_heap_1,
      system_monitor_long_schedule,
      system_monitor_large_heap_2, bad_flag, trace_delivered,
-     trap_exit_self_receive].
+     trap_exit_self_receive, trace_info_badarg, erl_704,
+     ms_excessive_nesting].
 
 init_per_testcase(_Case, Config) ->
     [{receiver,spawn(fun receiver/0)}|Config].
@@ -128,8 +130,7 @@ receive_trace(Config) when is_list(Config) ->
 		       {true, true}]),
 
     %% Remote messages
-    OtherName = atom_to_list(?MODULE)++"_receive_trace",
-    {ok, OtherNode} = start_node(OtherName),
+    {ok, Peer, OtherNode} = ?CT_PEER(),
     RemoteProc = spawn_link(OtherNode, ?MODULE, process, [self()]),
     io:format("RemoteProc = ~p ~n", [RemoteProc]),
 
@@ -177,7 +178,7 @@ receive_trace(Config) when is_list(Config) ->
     F2(NN, {true, true}),
 
     unlink(RemoteProc),
-    true = stop_node(OtherNode),
+    peer:stop(Peer),
 
     %% Timeout
     Receiver ! {set_timeout, 10},
@@ -645,8 +646,7 @@ procs_trace(Config) when is_list(Config) ->
 
 dist_procs_trace(Config) when is_list(Config) ->
     ct:timetrap({seconds, 15}),
-    OtherName = atom_to_list(?MODULE)++"_dist_procs_trace",
-    {ok, OtherNode} = start_node(OtherName),
+    {ok, Peer, OtherNode} = ?CT_PEER(),
     Self = self(),
     process_flag(trap_exit, true),
     %%
@@ -663,7 +663,7 @@ dist_procs_trace(Config) when is_list(Config) ->
     Proc1 ! {trap_exit_please, true},
     Proc3 = receive {spawned, Proc1, P3} -> P3 end,
     io:format("Proc3 = ~p ~n", [Proc3]),
-    {trace, Proc1, getting_linked, Proc3} = receive_first_trace(),
+    {trace, Proc1, link, Proc3} = receive_first_trace(),
     Reason3 = make_ref(),
     Proc1 ! {send_please, Proc3, {exit_please, Reason3}},
     receive {Proc1, {'EXIT', Proc3, Reason3}} -> ok end,
@@ -693,7 +693,7 @@ dist_procs_trace(Config) when is_list(Config) ->
 
     %%
     %% exit (with registered name, due to link)
-    Name = list_to_atom(OtherName),
+    Name = list_to_atom(hd(string:lexemes(atom_to_list(OtherNode), "@"))),
     Reason2 = make_ref(),
     Proc1 ! {link_please, Proc2},
     {trace, Proc1, link, Proc2} = receive_first_trace(),
@@ -706,8 +706,7 @@ dist_procs_trace(Config) when is_list(Config) ->
     receive_nothing(),
     %%
     %% Done.
-    true = stop_node(OtherNode),
-    ok.
+    peer:stop(Peer).
 
 %% Test trace(new, How, [procs]).
 procs_new_trace(Config) when is_list(Config) ->
@@ -957,15 +956,14 @@ do_system_monitor_long_schedule() ->
         {Self,L} when is_list(L) ->
             ok
     after 1000 ->
-              ct:fail(no_trace_of_pid)
+            ct:fail(no_trace_of_pid)
     end,
     "ok" = erlang:port_control(Port,1,[]),
-    "ok" = erlang:port_control(Port,2,[]),
     receive
         {Port,LL} when is_list(LL) ->
             ok
     after 1000 ->
-              ct:fail(no_trace_of_port)
+            ct:fail(no_trace_of_port)
     end,
     port_close(Port),
     erlang:system_monitor(undefined),
@@ -1683,6 +1681,18 @@ bad_flag(Config) when is_list(Config) ->
     {'EXIT', {badarg, _}} = (catch erlang:trace(new,
                                                 true,
                                                 [not_a_valid_flag])),
+
+    %% Leaks of {tracer,_} in OTP 23.2
+    Pid = spawn(fun() -> receive die -> ok end end),
+    1 = erlang:trace(Pid, true, [{tracer, self()},
+                                 {tracer, self()}]),
+    Pid ! die,
+    {'EXIT', {badarg, _}} =
+        (catch erlang:trace(new, true, [{tracer, self()}
+                                        | improper])),
+    {'EXIT', {badarg, _}} =
+        (catch erlang:trace(new, true, [{tracer, self()},
+                                        not_a_valid_flag])),
     ok.
 
 %% Test erlang:trace_delivered/1
@@ -1711,7 +1721,7 @@ trace_delivered(Config) when is_list(Config) ->
 
 %% This testcase checks that receive trace works on exit signal messages
 %% when the sender of the exit signal is the process itself.
-trap_exit_self_receive(Config) ->
+trap_exit_self_receive(Config) when is_list(Config) ->
     Parent = self(),
     Proc = spawn_link(fun() -> process(Parent) end),
 
@@ -1734,6 +1744,52 @@ trap_exit_self_receive(Config) ->
     receive_nothing(),
     ok.
 
+trace_info_badarg(Config) when is_list(Config) ->
+    catch erlang:trace_info({a,b,c},d),
+    ok.
+
+%% An incoming suspend monitor down wasn't handled
+%% correct when the local monitor half had been
+%% removed with an emulator crash as result.
+erl_704(Config) when is_list(Config) ->
+    erl_704_test(100).
+
+erl_704_test(0) ->
+    ok;
+erl_704_test(N) ->
+    P = spawn(fun () -> receive infinity -> ok end end),
+    erlang:suspend_process(P),
+    exit(P, kill),
+    (catch erlang:resume_process(P)),
+    erl_704_test(N-1).
+
+ms_excessive_nesting(Config) when is_list(Config) ->
+    MkMSCond = fun (_Fun, N) when N < 0 -> true;
+                   (Fun, N) -> {'or', {'=:=', N, '$1'}, Fun(Fun, N-1)}
+               end,
+    %% Ensure it compiles with substantial but reasonable
+    %% (hmm...) nesting
+    MS = [{['$1'], [MkMSCond(MkMSCond, 100)], []}],
+    io:format("~p~n", [erlang:match_spec_test([1], MS, trace)]),
+    _ = erlang:trace_pattern({?MODULE, '_', '_'}, MS, []),
+    %% Now test a match spec using excessive nesting. This
+    %% used to seg-fault the emulator due to recursion
+    %% beyond the end of the C-stack.
+    %%
+    %% We expect to get a system_limit error, but don't
+    %% fail if it compiles (someone must have rewritten
+    %% compilation of match specs to use an explicit
+    %% stack instead of using recursion).
+    ENMS = [{['$1'], [MkMSCond(MkMSCond, 1000000)], []}],
+    io:format("~p~n", [erlang:match_spec_test([1], ENMS, trace)]),
+    try
+        _ = erlang:trace_pattern({?MODULE, '_', '_'}, ENMS, []),
+        {comment, "compiled"}
+    catch
+        error:system_limit ->
+            {comment, "got system_limit"}
+    end.
+
 drop_trace_until_down(Proc, Mon) ->
     drop_trace_until_down(Proc, Mon, false, 0, 0).
 
@@ -1885,16 +1941,6 @@ fun_spawn(Fun, Args) ->
     spawn_link(erlang, apply, [Fun, Args]).
 
 
-start_node(Name) ->
-    Pa = filename:dirname(code:which(?MODULE)),
-    Cookie = atom_to_list(erlang:get_cookie()),
-    test_server:start_node(Name, slave,
-                           [{args, "-setcookie " ++ Cookie ++" -pa " ++ Pa}]).
-
-stop_node(Node) ->
-    test_server:stop_node(Node).
-
-
 wait_for_empty_runq(DeadLine) ->
     case statistics(run_queue) of
         0 -> true;
