@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2005-2020. All Rights Reserved.
+%% Copyright Ericsson AB 2005-2022. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -48,7 +48,8 @@
          exceptions_after_match_failure/1,
          bad_phi_paths/1,many_clauses/1,
          combine_empty_segments/1,hangs_forever/1,
-         bs_saved_position_units/1,empty_get_binary/1]).
+         bs_saved_position_units/1,empty_matches/1,
+         trim_bs_start_match_resume/1]).
 
 -export([coverage_id/1,coverage_external_ignore/2]).
 
@@ -87,7 +88,8 @@ groups() ->
        matching_meets_apply,bs_start_match2_defs,
        exceptions_after_match_failure,bad_phi_paths,
        many_clauses,combine_empty_segments,hangs_forever,
-       bs_saved_position_units,empty_get_binary]}].
+       bs_saved_position_units,empty_matches,
+       trim_bs_start_match_resume]}].
 
 init_per_suite(Config) ->
     test_lib:recompile(?MODULE),
@@ -221,6 +223,16 @@ int_float(Config) when is_list(Config) ->
     <<103133.0:64/float>> = <<103133:64/float>>,
     <<103133:64/float>> = <<103133:64/float>>,
 
+    %% Must work with type and flags in either order.
+    NativeFortyTwo = id(<<42/native-float>>),
+    <<42/float-native>> = NativeFortyTwo,
+    <<42/native-float>> = NativeFortyTwo,
+
+    %% Must work with redundant flags.
+    SixtyFour = id(<<64/float>>),
+    <<64/unsigned-float>> = SixtyFour,
+    <<64/float-unsigned>> = SixtyFour,
+
     %% Coverage of error cases in sys_pre_expand:coerce_to_float/2.
     case id(default) of
 	<<(1 bsl 1024):64/float>> ->
@@ -603,6 +615,10 @@ sum(<<>>, Last, Sum) -> {Sum,Last}.
 
 bin_and_float(Config) when is_list(Config) ->
     14.0 = bin_and_float(<<1.0/float,2.0/float,3.0/float>>, 0.0),
+
+    Sz = id(1),
+    <<>> = << <<0>> || <<1:Sz/float-unit:63>> <= <<2:30>> >>,
+
     ok.
 
 bin_and_float(<<X/float,Y/float,Z/float,T/binary>>, Sum) when is_float(X),
@@ -835,6 +851,11 @@ coverage(Config) when is_list(Config) ->
 
     {<<"abc">>,<<"tag">>} = coverage_trim_3([<<"abc","tag">>], 3),
 
+    <<57348:16/native>> = coverage_trim_4(<<"abc">>),
+
+    %% Cover code in beam_ssa_codegen.
+    ok = coverage_beam_ssa_codegen(<<2>>),
+
     ok.
 
 coverage_fold(Fun, Acc, <<H,T/binary>>) ->
@@ -963,9 +984,28 @@ coverage_trim_3(CipherTextFragment, TagLen) ->
         iolist_to_binary(CipherTextFragment),
     {CipherText, CipherTag}.
 
+coverage_trim_4(Bin) ->
+    << <<X:16>> || <<X:8>> <= Bin >>,
+    try
+        <<57348:16/native>>
+    catch
+        _:_ ->
+            error
+    end.
+
 printable_char($a) -> true;
 printable_char(_) -> false.
 
+coverage_beam_ssa_codegen(Bin) ->
+    %% With +r21 there will be a copy instruction that copies
+    %% a map literal to an x register, thus covering a line
+    %% beam_ssa_codegen:opt_allocate_defs/2.
+    case #{1 => 42} of
+        #{1 := 42} ->
+            << <<0>> || <<2>> <= Bin >>
+    end,
+    ok.
+
 multiple_uses(Config) when is_list(Config) ->
     {344,62879,345,<<245,159,1,89>>} = multiple_uses_1(<<1,88,245,159,1,89>>),
     true = multiple_uses_2(<<0,0,197,18>>),
@@ -1336,6 +1376,7 @@ bad_size(Config) when is_list(Config) ->
     Tuple = {a,b,c},
     Binary = <<1,2,3>>,
     Atom = an_atom,
+    NaN = <<(-1):32>>,
 
     {'EXIT',{{badmatch,<<>>},_}} = (catch <<32:Tuple>> = id(<<>>)),
     {'EXIT',{{badmatch,<<>>},_}} = (catch <<32:Binary>> = id(<<>>)),
@@ -1348,12 +1389,16 @@ bad_size(Config) when is_list(Config) ->
     {'EXIT',{{badmatch,<<>>},_}} = (catch <<42.0:Binary/float>> = id(<<>>)),
     {'EXIT',{{badmatch,<<>>},_}} = (catch <<42.0:Atom/float>> = id(<<>>)),
     {'EXIT',{{badmatch,<<>>},_}} = (catch <<42.0:2.5/float>> = id(<<>>)),
+    {'EXIT',{{badmatch,<<>>},_}} = (catch <<42.0:1/float>> = id(<<>>)),
+    {'EXIT',{{badmatch,NaN},_}} = (catch <<42.0:32/float>> = id(NaN)),
 
     %% Matched out value is ignored.
     {'EXIT',{{badmatch,<<>>},_}} = (catch <<_:Binary>> = id(<<>>)),
     {'EXIT',{{badmatch,<<>>},_}} = (catch <<_:Tuple>> = id(<<>>)),
     {'EXIT',{{badmatch,<<>>},_}} = (catch <<_:Atom>> = id(<<>>)),
     {'EXIT',{{badmatch,<<>>},_}} = (catch <<_:2.5>> = id(<<>>)),
+    {'EXIT',{{badmatch,<<1:1>>},_}} = (catch <<_:1/float>> = id(<<1:1>>)),
+    {'EXIT',{{badmatch,NaN},_}} = (catch <<_:32/float>> = id(NaN)),
 
     no_match = bad_all_size(<<>>),
     no_match = bad_all_size(<<1,2,3>>),
@@ -1446,13 +1491,11 @@ haystack_2(Haystack) ->
 fc({'EXIT',{function_clause,_}}) -> ok;
 fc({'EXIT',{{case_clause,_},_}}) when ?MODULE =:= bs_match_inline_SUITE -> ok.
 
-fc(Name, Args, {'EXIT',{function_clause,[{?MODULE,Name,Args,_}|_]}}) -> ok;
-fc(Name, Args, {'EXIT',{function_clause,[{?MODULE,Name,Arity,_}|_]}})
-  when length(Args) =:= Arity ->
-    true = test_server:is_native(?MODULE);
-fc(_, Args, {'EXIT',{{case_clause,ActualArgs},_}})
+fc(Name, Args, {'EXIT',{function_clause,[{?MODULE,Name,Args,_}|_]}}) ->
+    ok;
+fc(Name, Args, {'EXIT',{function_clause,[{?MODULE,_,Args,_}|_]}})
   when ?MODULE =:= bs_match_inline_SUITE ->
-    Args = tuple_to_list(ActualArgs).
+    ok.
 
 %% Cover the clause handling bs_context to binary in
 %% beam_block:initialized_regs/2.
@@ -1604,9 +1647,9 @@ mos_int_receive(Msg) ->
     Res.
 
 mos_int_fun(B) ->
-    L = ignore_me,
-    F = fun ([<<L,I:L,X:32>>]) -> {I,X};
-            ([<<L,I:L,X:64>>]) -> {I,X}
+    _L = ignore_me,
+    F = fun ([<<_L,I:_L,X:32>>]) -> {I,X};
+            ([<<_L,I:_L,X:64>>]) -> {I,X}
         end,
     F(B).
 
@@ -2391,14 +2434,23 @@ bspu_1(<<Bin/binary>> = Bin) ->
     [Chunk || <<Chunk:5/binary>> <= Bin].
 
 
-empty_get_binary(Config) when is_list(Config) ->
-    {<<>>, <<1,2,3,4:4>>} = egb_1(<<1,2,3,4:4>>),
-    {<<>>, <<1,2,3>>} = egb_1(<<1,2,3>>),
-    {<<>>, <<>>} = egb_1(<<>>),
+empty_matches(Config) when is_list(Config) ->
+    {<<>>, <<1,2,3,4:4>>} = em_1(<<1,2,3,4:4>>),
+    {<<>>, <<1,2,3>>} = em_1(<<1,2,3>>),
+    {<<>>, <<>>} = em_1(<<>>),
+
+    <<0,1,0,2,0,3>> = em_2(id(<<1,2,3>>)),
+    <<>> = em_2(id(<<>>)),
+
+    <<1,2,3>> = em_3(id(<<1,2,3>>)),
+    <<>> = em_3(id(<<>>)),
+
+    <<Zero:0/unit:1>> = id(<<>>),
+    0 = id(Zero),
 
     ok.
 
-egb_1(Bytes) ->
+em_1(Bytes) ->
     {Term, Bytes} = begin
                         <<V2@V0:0/binary-unit:8,V2@Buf1/bitstring>> = Bytes,
                         V2@Conv2 = binary:copy(V2@V0),
@@ -2406,6 +2458,30 @@ egb_1(Bytes) ->
                     end,
     {Term, Bytes}.
 
+em_2(Bin) ->
+    <<
+      <<K,N>> || <<K:0,N>> <= Bin
+    >>.
+
+%% The validator didn't agree with the type optimization pass on what the unit
+%% of empty binary segments should be.
+em_3(<<V:0/binary,Rest/bits>>) ->
+    em_3_1(V),
+    Rest.
+
+em_3_1(I) -> I.
+
+%% beam_trim would sometimes crash when bs_start_match4 had {atom,resume} as
+%% its fail label.
+trim_bs_start_match_resume(Config) when is_list(Config) ->
+    <<Context/binary>> = id(<<>>),
+    <<>> = trim_bs_start_match_resume_1(Context),
+    ok.
+
+trim_bs_start_match_resume_1(<<Context/binary>>) ->
+    _ = id(Context),
+    Context.
+
 id(I) -> I.
 
 expand_and_squeeze(Config) when is_list(Config) ->
