@@ -71,6 +71,7 @@
          nodename/0,
 	 protocol_childspecs/0,
 	 epmd_module/0,
+         get_state/0,
          dist_listen/0]).
 
 -export([disconnect/1, passive_cnct/1]).
@@ -195,6 +196,23 @@ longnames() ->                 request(longnames).
 
 nodename() ->                  request(nodename).
 
+-spec get_state() -> #{started => no | static | dynamic,
+                       name => atom(),
+                       name_type => static | dynamic,
+                       name_domain => shortnames | longnames}.
+get_state() ->
+    case whereis(net_kernel) of
+        undefined ->
+            case retry_request_maybe(get_state) of
+                ignored ->
+                    #{started => no};
+                Reply ->
+                    Reply
+            end;
+        _ ->
+            request(get_state)
+    end.
+
 -spec stop() -> ok | {error, Reason} when
       Reason :: not_allowed | not_found.
 stop() ->
@@ -561,7 +579,7 @@ do_auto_connect_2(Node, ConnId, From, State, ConnLookup) ->
                     {reply, false, State};
 
                 %% This might happen due to connection close
-                %% not beeing propagated to user space yet.
+                %% not being propagated to user space yet.
                 %% Save the day by just not connecting...
                 {ok, once} when ConnLookup =/= [],
                                 (hd(ConnLookup))#connection.state =:= up ->
@@ -833,6 +851,30 @@ handle_call({getopts, Node, Opts}, From, State) ->
     end,
     async_reply({reply, Return, State}, From);
 
+
+handle_call(get_state, From, State) ->
+    Started = case State#state.supervisor of
+                  net_sup -> static;
+                  _ -> dynamic
+              end,
+    {NameType,Name} = case {persistent_term:get(net_kernel, undefined), node()} of
+                          {undefined, Node} ->
+                              {static, Node};
+                          {dynamic_node_name, nonode@nohost} ->
+                              {dynamic, undefined};
+                          {dynamic_node_name, Node} ->
+                              {dynamic, Node}
+                      end,
+    NameDomain = case get(longnames) of
+                     true -> longnames;
+                     false -> shortnames
+                 end,
+    Return = #{started => Started,
+               name_type => NameType,
+               name => Name,
+               name_domain => NameDomain},
+    async_reply({reply, Return, State}, From);
+
 handle_call(_Msg, _From, State) ->
     {noreply, State}.
 
@@ -1047,6 +1089,23 @@ handle_info({SetupPid, {is_pending, Node}}, State) ->
     SetupPid ! {self(), {is_pending, Reply}},
     {noreply, State};
 
+handle_info({AcceptPid, {wait_pending, Node}}, State) ->
+    case get_conn(Node) of
+        {ok, #connection{state = up_pending,
+                         ctrlr = OldCtrlr,
+                         pending_owner = AcceptPid}} ->
+            %% Kill old controller to make sure new connection setup
+            %% does not get stuck.
+            ?debug({net_kernel, wait_pending, kill, OldCtrlr, new, AcceptPid}),
+            exit(OldCtrlr, wait_pending);
+        _ ->
+            %% Old connnection maybe already gone
+            ignore
+    end,
+    %% Exiting controller will trigger {Kernel,pending} reply
+    %% in up_pending_nodedown()
+    {noreply, State};
+
 %%
 %% Handle different types of process terminations.
 %%
