@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2007-2021. All Rights Reserved.
+%% Copyright Ericsson AB 2007-2022. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -141,11 +141,13 @@ init([Role, Sender, Host, Port, Socket, Options,  User, CbInfo]) ->
 	State1 = #state{static_env = #static_env{session_cache = Cache,
                                                  session_cache_cb = CacheCb
                                                 },
+                        connection_env = #connection_env{cert_key_alts = CertKeyAlts},
                         ssl_options = SslOptions,
                         session = Session0} = ssl_gen_statem:ssl_config(State0#state.ssl_options, Role, State0),
         State = case Role of
                     client ->
-                        Session = ssl_session:client_select_session({Host, Port, SslOptions}, Cache, CacheCb, Session0),
+                        CertKeyPairs = ssl_certificate:available_cert_key_pairs(CertKeyAlts),
+                        Session = ssl_session:client_select_session({Host, Port, SslOptions}, Cache, CacheCb, Session0, CertKeyPairs),
                         State1#state{session = Session};
                     server ->
                         State1
@@ -208,34 +210,36 @@ config_error(Type, Event, State) ->
 %%--------------------------------------------------------------------
 hello(internal, #client_hello{extensions = Extensions} = Hello, 
       #state{ssl_options = #{handshake := hello},
+             static_env = #static_env{role = server},
              handshake_env = HsEnv,
              start_or_recv_from = From} = State) ->
     {next_state, user_hello, State#state{start_or_recv_from = undefined,
                                          handshake_env = HsEnv#handshake_env{hello = Hello}},
      [{reply, From, {ok, Extensions}}]};
 hello(internal, #server_hello{extensions = Extensions} = Hello,
-      #state{ssl_options = #{
-                             handshake := hello},
+      #state{ssl_options = #{handshake := hello},
+             static_env = #static_env{role = client},
              handshake_env = HsEnv,
              start_or_recv_from = From} = State) ->   
     {next_state, user_hello,
      State#state{start_or_recv_from = undefined,
                  handshake_env = HsEnv#handshake_env{
                                    hello = Hello}}, [{reply, From, {ok, Extensions}}]};
-hello(internal, #client_hello{client_version = ClientVersion} = Hello, #state{ssl_options = SslOpts0,
-                                                                              connection_env = CEnv} = State0) ->
-    case choose_tls_fsm(SslOpts0, Hello) of
-        tls_1_3_fsm ->
-            %% Continue in TLS 1.3 'start' state
-            {next_state, start, State0, [{change_callback_module, tls_connection_1_3}, {next_event, internal, Hello}]};
-        tls_1_0_to_1_2_fsm ->
-            try handle_client_hello(Hello, State0) of
-                {ServerHelloExt, Type, State} ->
-                    {next_state, hello, State, [{next_event, internal, {common_client_hello, Type, ServerHelloExt}}]}
-            catch throw:#alert{} = Alert ->
-                    State = State0#state{connection_env = CEnv#connection_env{negotiated_version = ClientVersion}},
-                    ssl_gen_statem:handle_own_alert(Alert, hello, State)
-            end
+hello(internal, #client_hello{client_version = ClientVersion} = Hello,
+      #state{static_env = #static_env{role = server}, connection_env = CEnv} = State0) ->
+    try
+        #state{ssl_options = SslOpts} = State1 = tls_dtls_connection:handle_sni_extension(State0, Hello),
+        case choose_tls_fsm(SslOpts, Hello) of
+            tls_1_3_fsm ->
+                %% Continue in TLS 1.3 'start' state
+                {next_state, start, State1, [{change_callback_module, tls_connection_1_3}, {next_event, internal, Hello}]};
+            tls_1_0_to_1_2_fsm ->
+                {ServerHelloExt, Type, State} = handle_client_hello(Hello, State1),
+                {next_state, hello, State, [{next_event, internal, {common_client_hello, Type, ServerHelloExt}}]}
+        end
+    catch  throw:#alert{} = Alert ->
+            AlertState = State0#state{connection_env = CEnv#connection_env{negotiated_version = ClientVersion}},
+            ssl_gen_statem:handle_own_alert(Alert, hello, AlertState)
     end;
 hello(internal, #server_hello{} = Hello,
       #state{connection_states = ConnectionStates0,
@@ -343,22 +347,27 @@ connection(internal, #hello_request{},
 	   #state{static_env = #static_env{role = client,
                                            host = Host,
                                            port = Port,
+                                           cert_db = CertDbHandle,
+                                           cert_db_ref = CertDbRef,
                                            session_cache = Cache,
                                            session_cache_cb = CacheCb},
                   handshake_env = #handshake_env{
-                      renegotiation = {Renegotiation, peer},
-                      ocsp_stapling_state = OcspState},
-		  session = #session{own_certificates = OwnCerts} = Session0,
+                                     renegotiation = {Renegotiation, peer},
+                                     ocsp_stapling_state = OcspState},
+                  connection_env = #connection_env{cert_key_alts = CertKeyAlts},
+		  session = Session0,
 		  ssl_options = SslOpts, 
                   protocol_specific = #{sender := Pid},
 		  connection_states = ConnectionStates} = State0) ->
     try tls_sender:peer_renegotiate(Pid) of
         {ok, Write} ->
-            Session = ssl_session:client_select_session({Host, Port, SslOpts}, Cache, CacheCb, Session0),
+            CertKeyPairs = ssl_certificate:available_cert_key_pairs(CertKeyAlts),
+            Session = ssl_session:client_select_session({Host, Port, SslOpts}, Cache, CacheCb, Session0, CertKeyPairs),
             Hello = tls_handshake:client_hello(Host, Port, ConnectionStates, SslOpts,
                                                Session#session.session_id,
-                                               Renegotiation, OwnCerts, undefined,
-                                               undefined, maps:get(ocsp_nonce, OcspState, undefined)),
+                                               Renegotiation, undefined,
+                                               undefined, maps:get(ocsp_nonce, OcspState, undefined),
+                                               CertDbHandle, CertDbRef),
             {State, Actions} = tls_gen_connection:send_handshake(Hello, 
                                                                  State0#state{connection_states = 
                                                                                   ConnectionStates#{current_write => Write},
@@ -371,16 +380,19 @@ connection(internal, #hello_request{},
 connection(internal, #hello_request{},
 	   #state{static_env = #static_env{role = client,
                                            host = Host,
-                                           port = Port},
+                                           port = Port,
+                                           cert_db = CertDbHandle,
+                                           cert_db_ref = CertDbRef
+                                          },
                   handshake_env = #handshake_env{
                       renegotiation = {Renegotiation, _},
                       ocsp_stapling_state = OcspState},
-		  session = #session{own_certificates = OwnCerts},
 		  ssl_options = SslOpts, 
 		  connection_states = ConnectionStates} = State0) ->
     Hello = tls_handshake:client_hello(Host, Port, ConnectionStates, SslOpts,
-                                       <<>>, Renegotiation, OwnCerts, undefined,
-                                       undefined, maps:get(ocsp_nonce, OcspState, undefined)),
+                                       <<>>, Renegotiation, undefined,
+                                       undefined, maps:get(ocsp_nonce, OcspState, undefined),
+                                      CertDbHandle, CertDbRef),
 
     {State, Actions} = tls_gen_connection:send_handshake(Hello, State0),
     tls_gen_connection:next_event(hello, no_record, State, Actions);
@@ -434,7 +446,7 @@ terminate({shutdown, {sender_died, Reason}}, _StateName,
                                           transport_cb = Transport}} 
           = State) ->
     ssl_gen_statem:handle_trusted_certs_db(State),
-    tls_gen_connection:close(Reason, Socket, Transport, undefined, undefined);
+    tls_gen_connection:close(Reason, Socket, Transport, undefined);
 terminate(Reason, StateName, State) ->
     ssl_gen_statem:terminate(Reason, StateName, State).
 
@@ -449,11 +461,11 @@ code_change(_OldVsn, StateName, State, _) ->
 %%--------------------------------------------------------------------
 initial_state(Role, Sender, Host, Port, Socket, {SSLOptions, SocketOptions, Trackers}, User,
 	      {CbModule, DataTag, CloseTag, ErrorTag, PassiveTag}) ->
-    #{beast_mitigation := BeastMitigation,
-      erl_dist := IsErlDist,
+    #{erl_dist := IsErlDist,
       %% Use highest supported version for client/server random nonce generation
       versions := [Version|_],
       client_renegotiation := ClientRenegotiation} = SSLOptions,
+    BeastMitigation = maps:get(beast_mitigation, SSLOptions, disabled),
     ConnectionStates = tls_record:init_connection_states(Role,
                                                          Version,
                                                          BeastMitigation),
@@ -495,8 +507,7 @@ initial_state(Role, Sender, Host, Port, Socket, {SSLOptions, SocketOptions, Trac
                             }
       }.
 
-handle_client_hello(#client_hello{client_version = ClientVersion} = Hello, State0) ->
-    State = tls_dtls_connection:handle_sni_extension(State0, Hello),
+handle_client_hello(#client_hello{client_version = ClientVersion} = Hello, State) ->
     #state{connection_states = ConnectionStates0,
            static_env = #static_env{trackers = Trackers},
            handshake_env = #handshake_env{
@@ -504,8 +515,8 @@ handle_client_hello(#client_hello{client_version = ClientVersion} = Hello, State
                               renegotiation = {Renegotiation, _},
                               negotiated_protocol = CurrentProtocol,
                               sni_guided_cert_selection = SNICertSelection} = HsEnv,
-           connection_env = CEnv,
-           session = #session{own_certificates = OwnCerts} = Session0,
+           connection_env = #connection_env{cert_key_alts = CertKeyAlts} = CEnv,
+           session = Session0,
            ssl_options = SslOpts} = State,
     SessionTracker = proplists:get_value(session_id_tracker, Trackers),
     {Version, {Type, Session},
@@ -513,7 +524,7 @@ handle_client_hello(#client_hello{client_version = ClientVersion} = Hello, State
         tls_handshake:hello(Hello,
                             SslOpts,
                             {SessionTracker, Session0,
-                             ConnectionStates0, OwnCerts, KeyExAlg},
+                             ConnectionStates0, CertKeyAlts, KeyExAlg},
                             Renegotiation),
     Protocol = case Protocol0 of
                    undefined -> CurrentProtocol;
