@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2019-2019. All Rights Reserved.
+%% Copyright Ericsson AB 2019-2022. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -21,11 +21,56 @@
 %%
 -module(openssl_server_cert_SUITE).
 
-%% Note: This directive should only be used in test suites.
--compile(export_all).
 -include_lib("common_test/include/ct.hrl").
 -include_lib("public_key/include/public_key.hrl").
 
+%% Common test
+-export([all/0,
+         groups/0,
+         init_per_suite/1,
+         init_per_group/2,
+         init_per_testcase/2,
+         end_per_suite/1,
+         end_per_group/2,
+         end_per_testcase/2
+        ]).
+
+
+%% Test cases
+-export([no_auth/0,
+         no_auth/1,
+         auth/0,
+         auth/1,
+         client_auth_empty_cert_accepted/0,
+         client_auth_empty_cert_accepted/1,
+         client_auth_empty_cert_rejected/0,
+         client_auth_empty_cert_rejected/1,
+         client_auth_partial_chain/0,
+         client_auth_partial_chain/1,
+         client_auth_use_partial_chain/0,
+         client_auth_use_partial_chain/1,
+         client_auth_do_not_use_partial_chain/0,
+         client_auth_do_not_use_partial_chain/1,
+         client_auth_partial_chain_fun_fail/0,
+         client_auth_partial_chain_fun_fail/1,
+         missing_root_cert_no_auth/0,
+         missing_root_cert_no_auth/1,
+         unsupported_sign_algo_client_auth/0,
+         unsupported_sign_algo_client_auth/1,
+         unsupported_sign_algo_cert_client_auth/0,
+         unsupported_sign_algo_cert_client_auth/1,
+         hello_retry_request/0,
+         hello_retry_request/1,
+         custom_groups/0,
+         custom_groups/1,
+         hello_retry_client_auth/0,
+         hello_retry_client_auth/1,
+         hello_retry_client_auth_empty_cert_accepted/0,
+         hello_retry_client_auth_empty_cert_accepted/1,
+         hello_retry_client_auth_empty_cert_rejected/0,
+         hello_retry_client_auth_empty_cert_rejected/1
+        ]).
+
 %%--------------------------------------------------------------------
 %% Common Test interface functions -----------------------------------
 %%--------------------------------------------------------------------
@@ -52,20 +97,27 @@ groups() ->
      {rsa_pss_rsae, [], all_version_tests() ++ tls_1_3_tests()},
      {rsa_pss_pss, [], all_version_tests() ++ tls_1_3_tests()},
      {ecdsa_1_3, [], all_version_tests() ++ tls_1_3_tests()}
+     %% Enable test later when we have OpenSSL version that
+     %% is able to read EDDSA private key files
+     %%{eddsa_1_3, [], all_version_tests() ++ tls_1_3_tests()}
     ].
 
 protocol_groups() ->
-    [{group, 'tlsv1.3'},
-     {group, 'tlsv1.2'},
-     {group, 'tlsv1.1'},
-     {group, 'tlsv1'},
-     {group, 'dtlsv1.2'},
-     {group, 'dtlsv1'}
-     ].
-
-ssl_protocol_groups() ->
-    [{group, rsa},
-     {group, dsa}].
+    case ssl_test_lib:openssl_sane_dtls() of 
+        true ->
+            [{group, 'tlsv1.3'},
+             {group, 'tlsv1.2'},
+             {group, 'tlsv1.1'},
+             {group, 'tlsv1'},
+             {group, 'dtlsv1.2'},
+             {group, 'dtlsv1'}];
+        false ->
+            [{group, 'tlsv1.3'},
+             {group, 'tlsv1.2'},
+             {group, 'tlsv1.1'},
+             {group, 'tlsv1'}
+            ]
+    end. 
 
 pre_tls_1_3_protocol_groups() ->
     [{group, rsa},
@@ -76,7 +128,9 @@ tls_1_3_protocol_groups() ->
     [{group, rsa_1_3},
      {group, rsa_pss_rsae},
      {group, rsa_pss_pss},
-     {group, ecdsa_1_3}].
+     {group, ecdsa_1_3}
+     %%{group, eddsa_1_3}
+    ].
 
 tls_1_3_tests() ->
     [
@@ -92,7 +146,6 @@ all_version_tests() ->
      no_auth,
      auth,
      missing_root_cert_no_auth
-     %%invalid_signature_client
     ].
 
 init_per_suite(Config) ->
@@ -117,8 +170,8 @@ init_per_group(rsa = Group, Config0) ->
     Config = ssl_test_lib:make_rsa_cert(Config0),
     COpts = proplists:get_value(client_rsa_opts, Config),
     SOpts = proplists:get_value(server_rsa_opts, Config),
-    %% Make sure _rsa* suite is choosen by ssl_test_lib:start_server
-    Version = proplists:get_value(version,Config),
+    %% Make sure _rsa* suite is chosen by ssl_test_lib:start_server
+    Version = ssl_test_lib:protocol_version(Config),
     Ciphers = ssl_cert_tests:test_ciphers(fun(dhe_rsa) -> 
                                                   true;
                                              (ecdhe_rsa) -> 
@@ -141,8 +194,8 @@ init_per_group(rsa_1_3 = Group, Config0) ->
     Config = ssl_test_lib:make_rsa_cert(Config0),
     COpts = proplists:get_value(client_rsa_opts, Config),
     SOpts = proplists:get_value(server_rsa_opts, Config),
-    %% Make sure _rsa* suite is choosen by ssl_test_lib:start_server
-    Version = proplists:get_value(version,Config),
+    %% Make sure _rsa* suite is chosen by ssl_test_lib:start_server
+    Version = ssl_test_lib:protocol_version(Config),
     Ciphers = ssl_cert_tests:test_ciphers(undefined, Version),
     case Ciphers of
         [_|_] ->
@@ -180,13 +233,15 @@ init_per_group(Alg, Config) when Alg == rsa_pss_rsae;
 init_per_group(ecdsa = Group, Config0) ->
     PKAlg = crypto:supports(public_keys),
     case lists:member(ecdsa, PKAlg) andalso (lists:member(ecdh, PKAlg) orelse 
-                                             lists:member(dh, PKAlg)) of
+                                                lists:member(dh, PKAlg)) 
+        andalso (ssl_test_lib:openssl_ecdsa_suites() =/= []) 
+    of
         true ->
             Config = ssl_test_lib:make_ecdsa_cert(Config0),
             COpts = proplists:get_value(client_ecdsa_opts, Config),
             SOpts = proplists:get_value(server_ecdsa_opts, Config),
-            %% Make sure ecdh* suite is choosen by ssl_test_lib:start_server
-            Version = proplists:get_value(version,Config),
+            %% Make sure ecdh* suite is chosen by ssl_test_lib:start_server
+            Version = ssl_test_lib:protocol_version(Config),
             Ciphers =  ssl_cert_tests:test_ciphers(fun(ecdh_ecdsa) -> 
                                                            true;
                                                       (ecdhe_ecdsa) -> 
@@ -212,13 +267,15 @@ init_per_group(ecdsa = Group, Config0) ->
 init_per_group(ecdsa_1_3 = Group, Config0) ->
     PKAlg = crypto:supports(public_keys),
     case lists:member(ecdsa, PKAlg) andalso (lists:member(ecdh, PKAlg) orelse
-                                             lists:member(dh, PKAlg)) of
+                                             lists:member(dh, PKAlg)) 
+        andalso (ssl_test_lib:openssl_ecdsa_suites() =/= []) 
+    of
         true ->
             Config = ssl_test_lib:make_ecdsa_cert(Config0),
             COpts = proplists:get_value(client_ecdsa_opts, Config),
             SOpts = proplists:get_value(server_ecdsa_opts, Config),
-            %% Make sure ecdh* suite is choosen by ssl_test_lib:start_server
-            Version = proplists:get_value(version,Config),
+            %% Make sure ecdh* suite is chosen by ssl_test_lib:start_server
+            Version = ssl_test_lib:protocol_version(Config),
             Ciphers =  ssl_cert_tests:test_ciphers(undefined, Version),
             case Ciphers of
                 [_|_] ->
@@ -235,15 +292,42 @@ init_per_group(ecdsa_1_3 = Group, Config0) ->
         false ->
             {skip, "Missing EC crypto support"}
     end;
-init_per_group(Group, Config0) when Group == dsa ->
+
+init_per_group(eddsa_1_3, Config0) ->
     PKAlg = crypto:supports(public_keys),
-    case lists:member(dss, PKAlg) andalso lists:member(dh, PKAlg) of
+    PrivDir = proplists:get_value(priv_dir, Config0),
+    case lists:member(eddsa, PKAlg) andalso (lists:member(ecdh, PKAlg)) of
+        true ->
+            Conf = public_key:pkix_test_data(#{server_chain => #{root => ssl_test_lib:eddsa_conf(),
+                                                                 intermediates => [ssl_test_lib:eddsa_conf()],
+                                                                 peer =>  ssl_test_lib:eddsa_conf()},
+                                               client_chain => #{root => ssl_test_lib:eddsa_conf(),
+                                                                 intermediates => [ssl_test_lib:eddsa_conf()],
+                                                                 peer =>  ssl_test_lib:eddsa_conf()}}),
+            [{server_config, SOpts},
+             {client_config, COpts}] = x509_test:gen_pem_config_files(Conf, filename:join(PrivDir, "client_eddsa"),
+                                                                      filename:join(PrivDir, "server_eddsa")),
+
+            [{cert_key_alg, eddsa} |
+             lists:delete(cert_key_alg,
+                          [{client_cert_opts, COpts},
+                           {server_cert_opts, SOpts} |
+                           lists:delete(server_cert_opts,
+                                        lists:delete(client_cert_opts, Config0))]
+                         )];
+        false ->
+            {skip, "Missing EC crypto support"}
+    end;
+init_per_group(dsa = Group, Config0) ->
+    PKAlg = crypto:supports(public_keys),
+    case lists:member(dss, PKAlg) andalso lists:member(dh, PKAlg)  andalso 
+        (ssl_test_lib:openssl_dsa_suites() =/= [])  of
         true ->
             Config = ssl_test_lib:make_dsa_cert(Config0),    
             COpts = proplists:get_value(client_dsa_opts, Config),
             SOpts = proplists:get_value(server_dsa_opts, Config),
-            %% Make sure dhe_dss* suite is choosen by ssl_test_lib:start_server
-            Version = proplists:get_value(version,Config),
+            %% Make sure dhe_dss* suite is chosen by ssl_test_lib:start_server
+            Version = ssl_test_lib:protocol_version(Config),
             Ciphers =  ssl_cert_tests:test_ciphers(fun(dh_dss) -> 
                                                            true;
                                                       (dhe_dss) -> 
@@ -310,15 +394,15 @@ client_auth_partial_chain(Config) when is_list(Config) ->
     ssl_cert_tests:client_auth_partial_chain(Config).
 
 %%--------------------------------------------------------------------
-client_auth_allow_partial_chain() ->
-    ssl_cert_tests:client_auth_allow_partial_chain().
-client_auth_allow_partial_chain(Config) when is_list(Config) ->
-    ssl_cert_tests:client_auth_allow_partial_chain(Config).
+client_auth_use_partial_chain() ->
+    ssl_cert_tests:client_auth_use_partial_chain().
+client_auth_use_partial_chain(Config) when is_list(Config) ->
+    ssl_cert_tests:client_auth_use_partial_chain(Config).
 %%--------------------------------------------------------------------
-client_auth_do_not_allow_partial_chain() ->
-   ssl_cert_tests:client_auth_do_not_allow_partial_chain().
-client_auth_do_not_allow_partial_chain(Config) when is_list(Config) ->
-    ssl_cert_tests:client_auth_do_not_allow_partial_chain(Config).
+client_auth_do_not_use_partial_chain() ->
+   ssl_cert_tests:client_auth_do_not_use_partial_chain().
+client_auth_do_not_use_partial_chain(Config) when is_list(Config) ->
+    ssl_cert_tests:client_auth_do_not_use_partial_chain(Config).
 
 %%--------------------------------------------------------------------
 client_auth_partial_chain_fun_fail() ->
@@ -332,16 +416,6 @@ missing_root_cert_no_auth() ->
 missing_root_cert_no_auth(Config) when is_list(Config) ->
     ssl_cert_tests:missing_root_cert_no_auth(Config).
 
-%%--------------------------------------------------------------------
-invalid_signature_client() ->
-    ssl_cert_tests:invalid_signature_client().
-invalid_signature_client(Config) when is_list(Config) ->
-    ssl_cert_tests:invalid_signature_client(Config).
-%%--------------------------------------------------------------------
-invalid_signature_server() ->
-    ssl_cert_tests:invalid_signature_client().
-invalid_signature_server(Config) when is_list(Config) ->
-    ssl_cert_tests:invalid_signature_client(Config).
 
 %%--------------------------------------------------------------------
 %% TLS 1.3 Test Cases ------------------------------------------------
